import types
import packaging.version
import sys
import os
import importlib


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


class Version(packaging.version.Version):
    '''Extended implementation of packaging.version.Version'''

    def match(self, condition: str) -> bool:
        '''
        Determine whether the given condition are met
        Args:
            condition(str) : conditions for judgment
        Returns:
            bool: True if the given version condition are met, else False
        Examples:
            .. code-block:: python
                Version('1.2.0').match('>=1.2.0a')
        '''
        if not condition:
            return True
        if condition.startswith('>='):
            version = condition[2:]
            _comp = self.__ge__
        elif condition.startswith('>'):
            version = condition[1:]
            _comp = self.__gt__
        elif condition.startswith('<='):
            version = condition[2:]
            _comp = self.__le__
        elif condition.startswith('<'):
            version = condition[1:]
            _comp = self.__lt__
        elif condition.startswith('=='):
            version = condition[2:]
            _comp = self.__eq__
        elif condition.startswith('='):
            version = condition[1:]
            _comp = self.__eq__
        else:
            version = condition
            _comp = self.__eq__

        return _comp(Version(version))

    def __lt__(self, other):
        if isinstance(other, str):
            other = Version(other)
        return super().__lt__(other)

    def __le__(self, other):
        if isinstance(other, str):
            other = Version(other)
        return super().__le__(other)

    def __gt__(self, other):
        if isinstance(other, str):
            other = Version(other)
        return super().__gt__(other)

    def __ge__(self, other):
        if isinstance(other, str):
            other = Version(other)
        return super().__ge__(other)

    def __eq__(self, other):
        if isinstance(other, str):
            other = Version(other)
        return super().__eq__(other)


def load_py_dir(python_dir: str):
    model_basename = os.path.basename(python_dir)
    spec = importlib.util.spec_from_file_location(
        model_basename, os.path.join(python_dir, '__init__.py'))
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


def load_py_module(python_path: str, py_module_name: str) -> types.ModuleType:
    '''
    Load the specified python module.

    Args:
        python_path(str) : The directory where the python module is located
        py_module_name(str) : Module name to be loaded
    '''
    sys.path.insert(0, python_path)

    # Delete the cache module to avoid hazards. For example, when the user reinstalls a HubModule,
    # if the cache is not cleared, then what the user gets at this time is actually the HubModule
    # before uninstallation, this can cause some strange problems, e.g, fail to load model parameters.
    if py_module_name in sys.modules:
        sys.modules.pop(py_module_name)

    py_module = importlib.import_module(py_module_name)
    sys.path.pop(0)

    return py_module


# 导入hashlib和threading模块
import hashlib
import threading

# 定义一个FileMonitor类
class FileMonitor:
    # 构造函数，初始化属性
    def __init__(self, filename, callback=[]):
        # 要检测的文件名
        self.filename = filename
        # 文件的初始哈希值
        self.hash_value = self.get_hash(filename)
        # 文件被修改时要调用的函数
        self.callback = callback

    def appendCB(self, cb):
        self.callback.append(cb)

    # 计算文件的哈希值
    def get_hash(self, filename):
        # 打开文件，以二进制模式读取
        with open(filename, "rb") as f:
            # 创建一个md5对象
            md5 = hashlib.md5()
            # 循环读取文件内容，更新md5对象
            while True:
                data = f.read(1024)
                if not data:
                    break
                md5.update(data)
            # 返回文件的哈希值
            return md5.hexdigest()

    # 检测文件的哈希值，并在文件被修改时调用callback函数
    def check_file(self):
        # 循环检测文件的哈希值
        while True:
            try:
                # 获取文件的当前哈希值
                new_hash_value = self.get_hash(self.filename)
                # 如果当前哈希值和初始哈希值不同，说明文件被修改了
                if new_hash_value != self.hash_value:
                    # 调用callback函数
                    if self.callback is not None:
                        if isinstance(self.callback, list):
                            for c in self.callback:
                                c(self.filename)
                        else:
                            self.callback(self.filename)
                    # 更新初始哈希值为当前哈希值
                    self.hash_value = new_hash_value
            except:
                pass

    # 创建一个后台线程，执行check_file方法
    def start(self):
        # 创建一个后台线程，传入check_file方法作为目标函数（target）
        t = threading.Thread(target=self.check_file)
        # 设置线程为守护线程（daemon），这样主程序结束时，线程也会自动结束
        t.setDaemon(True)
        # 启动线程
        t.start()





import torch
import torch.nn as nn
try:
    from encoder import TransformerEncoder
    from decoder import TransformerDecoder
    from embedding import Embeddings, PADDING_INDEX, _position_encoding_default, _position_encoding_absolute
    from multi_headed_attn import MultiHeadedAttention, ATTEN_TYPE_SELF, ATTEN_TYPE_HYGRA
except:
    from aicmder.torch.transformer import TransformerEncoder, TransformerDecoder, Embeddings, PADDING_INDEX, ATTEN_TYPE_SELF, _position_encoding_default, _position_encoding_absolute
from aicmder.torch.dist_model import Trainable
from torch.nn import functional as F
from torch import Tensor
from typing import Optional, Any, Union, Callable
from torch.nn.init import xavier_uniform_


_attn_type = ATTEN_TYPE_SELF
# _attn_type = ATTEN_TYPE_HYGRA 
class TransformerSample(nn.Module):

    def __init__(self, args):

        super(TransformerSample, self).__init__()

        if "src_word_vocab_size" in args:
            source_embeddings = Embeddings(args.d_model, args.src_word_vocab_size, PADDING_INDEX, position_encoding=_position_encoding_default)
        else:
            source_embeddings = None

        num_head = args.num_head
        d_ff = args.d_ff
        num_layer = args.nlayers
        self.encoder = TransformerEncoder(
            num_head, args.d_model, args.dropout, args.max_relative_positions, d_ff, embeddings=source_embeddings, num_layers=num_layer, attn_type=_attn_type)
       
        if "tgt_word_vocab_size" in args: 
            target_embeddings = Embeddings(args.d_model, args.tgt_word_vocab_size, PADDING_INDEX, position_encoding=_position_encoding_default) 
        else:
            target_embeddings = None
            
        self.decoder = TransformerDecoder(
            num_layers=num_layer,
            d_model=args.d_model,
            heads=num_head,
            d_ff=d_ff,
            copy_attn=False,
            embeddings=target_embeddings,
            dropout=args.dropout
        )

        # layer_norm_eps = 1e-5
        # batch_first = False
        # norm_first = False
        # factory_kwargs = {'device': None, 'dtype': None}
        # activation = F.relu
        # decoder_layer = nn.TransformerDecoderLayer(args.d_model, num_head, d_ff, args.dropout,
        #                                         activation, layer_norm_eps, batch_first, norm_first,
        #                                         **factory_kwargs)
        # decoder_norm = nn.LayerNorm(args.d_model, eps=layer_norm_eps, **factory_kwargs)
        # self.decoder = nn.TransformerDecoder(decoder_layer, num_layer, decoder_norm)
        
        self._reset_parameters()



    # def forward(self, src_input, tgt: Tensor, src_mask: Optional[Tensor] = None, tgt_mask: Optional[Tensor] = None,
    #             memory_mask: Optional[Tensor] = None, src_key_padding_mask: Optional[Tensor] = None,
    #             tgt_key_padding_mask: Optional[Tensor] = None, memory_key_padding_mask: Optional[Tensor] = None) -> Tensor:

    def forward(self, src_input, tgt_input, src_mask=None, tgt_pad_mask=None):
        memory_bank = self.encoder(src_input.float(), mask=src_mask)
        max_mem_len = memory_bank[0].shape[1] \
            if isinstance(memory_bank, list) else memory_bank.shape[1]
        memory_len = torch.tensor([src_input.size(1) for i in range(src_input.size(0))], device=memory_bank.device, dtype=torch.long)
        self.decoder.init_state(memory_len, max_mem_len)
        dec_outs, attns = self.decoder(tgt_input.float(), memory_bank=memory_bank, src_pad_mask=src_mask, tgt_pad_mask=tgt_pad_mask, attn_type=_attn_type)
        return dec_outs, attns


        # memory_bank = self.encoder(src_input.float())
        # output = self.decoder(tgt, memory_bank, tgt_mask=tgt_mask, memory_mask=memory_mask,
        #                       tgt_key_padding_mask=tgt_key_padding_mask,
        #                       memory_key_padding_mask=memory_key_padding_mask)
        # return output


    def _reset_parameters(self):
        r"""Initiate parameters in the transformer model."""

        for p in self.parameters():
            if p.dim() > 1:
                xavier_uniform_(p)

if __name__ == "__main__":
    trainable = Trainable(bs=32)
    args = dict()
    args["d_model"] = 512
    # args["src_word_vocab_size"] = 184866
    # args["tgt_word_vocab_size"] = 184866 
    args["num_head"] = 8
    args["dropout"] = 0.2
    args["d_ff"] = 2048
    args["max_relative_positions"] = 32
    args["nlayers"] = 1
    from aicmder.utils import AttrDict
    model = TransformerSample(AttrDict(args))
    trainable.on_pretrain_routine_start(model, verbose=True)
    





import os
from pathlib import Path
import re
from datetime import datetime
import contextlib
import glob
import inspect
import platform
import time
import signal
import threading
from typing import Optional
from multiprocessing import Pool
from tqdm import tqdm
from functools import partial


def is_writeable(dir, test=False):
    # Return True if directory has write permissions, test opening a file with write permissions if test=True
    if not test:
        return os.access(dir, os.R_OK)  # possible issues on Windows
    file = Path(dir) / 'tmp.txt'
    try:
        with open(file, 'w'):  # open file with write permissions
            pass
        file.unlink()  # remove file
        return True
    except OSError:
        return False


def is_pip():
    # Is file in a pip package?
    return 'site-packages' in Path(__file__).resolve().parts


def is_ascii(s=''):
    # Is string composed of all ASCII (no UTF) characters? (note str().isascii() introduced in python 3.7)
    s = str(s)  # convert list, tuple, None, etc. to str
    return len(s.encode().decode('ascii', 'ignore')) == len(s)


def is_chinese(s='人工智能'):
    # Is string composed of any Chinese characters?
    return re.search('[\u4e00-\u9fff]', s)


def file_age(path=__file__):
    # Return days since last file update
    dt = (datetime.now() - datetime.fromtimestamp(Path(path).stat().st_mtime))  # delta
    return dt.days  # + dt.seconds / 86400  # fractional days


def file_date(path=__file__):
    # Return human-readable file modification date, i.e. '2021-3-26'
    t = datetime.fromtimestamp(Path(path).stat().st_mtime)
    return f'{t.year}-{t.month}-{t.day}'


def file_size(path):
    # Return file/dir size (MB)
    mb = 1 << 20  # bytes to MiB (1024 ** 2)
    path = Path(path)
    if path.is_file():
        return path.stat().st_size / mb
    elif path.is_dir():
        return sum(f.stat().st_size for f in path.glob('**/*') if f.is_file()) / mb
    else:
        return 0.0


class Profile(contextlib.ContextDecorator):
    # Usage: @Profile() decorator or 'with Profile():' context manager
    def __enter__(self):
        self.start = time.time()

    def __exit__(self, type, value, traceback):
        print(f'Profile results: {time.time() - self.start:.5f}s')


class Timeout(contextlib.ContextDecorator):
    # Usage: @Timeout(seconds) decorator or 'with Timeout(seconds):' context manager
    def __init__(self, seconds, *, timeout_msg='', suppress_timeout_errors=True):
        self.seconds = int(seconds)
        self.timeout_message = timeout_msg
        self.suppress = bool(suppress_timeout_errors)

    def _timeout_handler(self, signum, frame):
        raise TimeoutError(self.timeout_message)

    def __enter__(self):
        if platform.system() != 'Windows':  # not supported on Windows
            signal.signal(signal.SIGALRM, self._timeout_handler)  # Set handler for SIGALRM
            signal.alarm(self.seconds)  # start countdown for SIGALRM to be raised

    def __exit__(self, exc_type, exc_val, exc_tb):
        if platform.system() != 'Windows':
            signal.alarm(0)  # Cancel SIGALRM if it's scheduled
            if self.suppress and exc_type is TimeoutError:  # Suppress TimeoutError
                return True


class WorkingDirectory(contextlib.ContextDecorator):
    # Usage: @WorkingDirectory(dir) decorator or 'with WorkingDirectory(dir):' context manager
    def __init__(self, new_dir):
        self.dir = new_dir  # new dir
        self.cwd = Path.cwd().resolve()  # current dir

    def __enter__(self):
        os.chdir(self.dir)

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self.cwd)


def try_except(func):
    # try-except function. Usage: @try_except decorator
    def handler(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Exception as e:
            print(e)

    return handler


def threaded(func):
    # Multi-threads a target function and returns thread. Usage: @threaded decorator
    def wrapper(*args, **kwargs):
        thread = threading.Thread(target=func, args=args, kwargs=kwargs, daemon=True)
        thread.start()
        return thread

    return wrapper


def multiple_threaded(func, param, loops, cpu_count=os.cpu_count()):
    f = partial(func, param)
    rets = []
    if isinstance(loops, zip):
        loops = list(loops)
    loop_len = len(loops)
    # print(loop_len, loops, list(loops))
    with Pool(cpu_count) as pool:
        for ret in tqdm(
            pool.imap(
                f,
                loops,
            ),
            total=loop_len,
        ):
            rets.append(ret)
    return rets

################## example ##############


def test(*p):
    print(p, len(p))
    return p[len(p) - 1]


if __name__ == "__main__":
    param = {}
    param["test"] = 11
    # rets = multiple_threaded(test, param, zip(range(2, 10), range(22, 30)))
    rets = multiple_threaded(test, param, range(2, 10))
    print(rets)


def methods(instance):
    # Get class/instance methods
    return [f for f in dir(instance) if callable(getattr(instance, f)) and not f.startswith("__")]


def colorstr(*input):
    # Colors a string https://en.wikipedia.org/wiki/ANSI_escape_code, i.e.  colorstr('blue', 'hello world')
    *args, string = input if len(input) > 1 else ('blue', 'bold', input[0])  # color arguments, string
    colors = {
        'black': '\033[30m',  # basic colors
        'red': '\033[31m',
        'green': '\033[32m',
        'yellow': '\033[33m',
        'blue': '\033[34m',
        'magenta': '\033[35m',
        'cyan': '\033[36m',
        'white': '\033[37m',
        'bright_black': '\033[90m',  # bright colors
        'bright_red': '\033[91m',
        'bright_green': '\033[92m',
        'bright_yellow': '\033[93m',
        'bright_blue': '\033[94m',
        'bright_magenta': '\033[95m',
        'bright_cyan': '\033[96m',
        'bright_white': '\033[97m',
        'end': '\033[0m',  # misc
        'bold': '\033[1m',
        'underline': '\033[4m'}
    return ''.join(colors[x] for x in args) + f'{string}' + colors['end']


def print_args(args: Optional[dict] = None, show_file=True, show_fcn=False):
    # Print function arguments (optional args dict)
    x = inspect.currentframe().f_back  # previous frame
    file, _, fcn, _, _ = inspect.getframeinfo(x)
    if args is None:  # get args automatically
        args, _, _, frm = inspect.getargvalues(x)
        args = {k: v for k, v in frm.items() if k in args}
    s = (f'{Path(file).stem}: ' if show_file else '') + (f'{fcn}: ' if show_fcn else '')
    print(colorstr(s) + ', '.join(f'{k}={v}' for k, v in args.items()))



# coding:utf-8
import sys

__version__ = '0.5.9'

# from aicmder.utils import utils
# sys.modules['aicmder.common.utils'] = utils
from aicmder.commands import *

# torch
from aicmder.torch import select_device, CudaDecorators, cpu, gpu, try_gpu, train_params, HyperParameters

# util
from aicmder.utils import WorkingDirectory, try_except, threaded, iSummaryWriter, FileMonitor

# benchmark
from aicmder.benchmark import model_info, torch_info

# module
from aicmder.module.module import Module
from aicmder.module.define import ModuleDefine
from aicmder.module.module import serving, runnable

# serving
from aicmder.service.worker import Worker
try:
    from aicmder.service.http_service import HTTPProxy
except Exception as e:
    print(e)
from aicmder.service.server import ServerQueue


from aicmder.service.PPworker import PPworker

# common
import aicmder.common as Common

try:
    from aicmder.service.http_service import Client
except Exception as e:
    print(e)


import sys
import os


def import_parent(current_file):
    # getting the name of the directory
    # where the this file is present.
    current = os.path.dirname(os.path.realpath(current_file))

    # Getting the parent directory name
    # where the current directory is present.
    parent = os.path.dirname(current)

    # adding the parent directory to
    # the sys.path.
    sys.path.append(parent)
    # print(parent, __file__)



def add_to_class(Class):
    """Defined in :numref:`sec_oo-design`"""
    def wrapper(obj):
        setattr(Class, obj.__name__, obj)
    return wrapper

from aicmder.utils import AttrDict



# src: https://github.com/OpenNMT/OpenNMT-py/blob/master/onmt/encoders/rnn_encoder.py
"""Define RNN-based encoders."""
from __future__ import division
import warnings
warnings.filterwarnings("ignore")
import torch
import torch.nn as nn
import torch.nn.functional as F

from aicmder.torch.transformer.base_model import EncoderBase
from torch.nn.utils.rnn import pack_padded_sequence as pack
from torch.nn.utils.rnn import pad_packed_sequence as unpack
from aicmder.torch.transformer.misc import aeq
from aicmder.torch import HyperParameters, Module
from aicmder.torch.transformer.attention import DotProductAttention

class RNNEncoder(EncoderBase):
    """ A generic recurrent neural network encoder.
    Args:
       rnn_type (:obj:`str`):
          style of recurrent unit to use, one of [RNN, LSTM, GRU, SRU]
       bidirectional (bool) : use a bidirectional RNN
       num_layers (int) : number of stacked layers
       hidden_size (int) : hidden size of each layer
       dropout (float) : dropout value for :obj:`nn.Dropout`
    """

    def __init__(self,
                 hidden_size=100,
                 num_layers=6,
                 rnn_type="LSTM",
                 bidirectional=True,
                 input_size=None,
                 embeddings=None,
                 dropout=0.2,
                 use_bridge=False,
                 use_last=True):
        super(RNNEncoder, self).__init__()

        num_directions = 2 if bidirectional else 1
        assert hidden_size % num_directions == 0
        self.hidden_size = hidden_size
        hidden_size = hidden_size // num_directions

        self.embeddings = embeddings
        assert embeddings is not None or input_size is not None
        if input_size is None:
            input_size = embeddings.embedding_size

        # Saves preferences for layer
        self.nlayers = num_layers
        self.use_last = use_last

        self.rnns = nn.ModuleList()
        for i in range(self.nlayers):
            input_size = input_size if i == 0 else hidden_size * num_directions
            kwargs = {'input_size': input_size,
                      'hidden_size': hidden_size,
                      'num_layers': 1,
                      'bidirectional': bidirectional,
                      'batch_first': True}
            rnn = getattr(nn, rnn_type)(**kwargs)
            self.rnns.append(rnn)

        self.dropout = nn.Dropout(dropout)
        # Initialize the bridge layer
        self.use_bridge = use_bridge
        if self.use_bridge:
            nl = 1 if self.use_last else num_layers
            self._initialize_bridge(rnn_type, hidden_size, nl)

    def count_parameters(self):
        params = list(self.rnns.parameters())
        if self.use_bridge:
            params = params + list(self.bridge.parameters())
        return sum(p.numel() for p in params if p.requires_grad)

    def forward(self, emb, lengths=None):
        "See :obj:`EncoderBase.forward()`"
        self._check_args(emb, lengths)

        if self.embeddings is not None:
            emb = self.embeddings(emb)

        packed_emb = emb
        if lengths is not None:
            # Lengths data is wrapped inside a Tensor.
            lengths, indices = torch.sort(lengths, 0, True)  # Sort by length (keep idx)
            packed_emb = pack(packed_emb[indices], lengths.tolist(), batch_first=True)
            _, _indices = torch.sort(indices, 0)  # Un-sort by length

        memory_bank, encoder_final = [], {'h_n': [], 'c_n': []}
        for i in range(self.nlayers):
            if i != 0:
                packed_emb = self.dropout(packed_emb)
                if lengths is not None:
                    packed_emb = pack(packed_emb, lengths.tolist(), batch_first=True)

            packed_emb, states = self.rnns[i](packed_emb)
            if isinstance(states, tuple):
                h_n, c_n = states
                encoder_final['c_n'].append(c_n)
            else:
                h_n = states
            encoder_final['h_n'].append(h_n)

            packed_emb = unpack(packed_emb, batch_first=True)[0] if lengths is not None else packed_emb
            if not self.use_last or i == self.nlayers - 1:
                memory_bank += [packed_emb[_indices]] if lengths is not None else [packed_emb]

        assert len(encoder_final['h_n']) != 0
        if self.use_last:
            memory_bank = memory_bank[-1]
            if len(encoder_final['c_n']) == 0:
                encoder_final = encoder_final['h_n'][-1]
            else:
                encoder_final = encoder_final['h_n'][-1], encoder_final['c_n'][-1]
        else:
            memory_bank = torch.cat(memory_bank, dim=2)
            if len(encoder_final['c_n']) == 0:
                encoder_final = torch.cat(encoder_final['h_n'], dim=0)
            else:
                encoder_final = torch.cat(encoder_final['h_n'], dim=0), \
                    torch.cat(encoder_final['c_n'], dim=0)

        if self.use_bridge:
            encoder_final = self._bridge(encoder_final)

        # TODO: Temporary hack is adopted to compatible with DataParallel
        # reference: https://github.com/pytorch/pytorch/issues/1591
        if memory_bank.size(1) < emb.size(1):
            dummy_tensor = torch.zeros(memory_bank.size(0),
                                       emb.size(1) - memory_bank.size(1),
                                       memory_bank.size(2)).type_as(memory_bank)
            memory_bank = torch.cat([memory_bank, dummy_tensor], 1)

        return encoder_final, memory_bank

    def _initialize_bridge(self,
                           rnn_type,
                           hidden_size,
                           num_layers):

        # LSTM has hidden and cell state, other only one
        number_of_states = 2 if rnn_type == "LSTM" else 1
        # Total number of states
        self.total_hidden_dim = hidden_size * num_layers

        # Build a linear layer for each
        self.bridge = nn.ModuleList([nn.Linear(self.total_hidden_dim,
                                               self.total_hidden_dim,
                                               bias=True)
                                     for _ in range(number_of_states)])

    def _bridge(self, hidden):
        """
        Forward hidden state through bridge
        """

        def bottle_hidden(linear, states):
            """
            Transform from 3D to 2D, apply linear and return initial size
            """
            size = states.size()
            result = linear(states.view(-1, self.total_hidden_dim))
            return F.relu(result).view(size)

        if isinstance(hidden, tuple):  # LSTM
            outs = tuple([bottle_hidden(layer, hidden[ix])
                          for ix, layer in enumerate(self.bridge)])
        else:
            outs = bottle_hidden(self.bridge[0], hidden)
        return outs

    def update_dropout(self, dropout):
        self.dropout.p = dropout

################### add attention
class AttnRNNEncoder(Module):

    def __init__(self, *args, use_attn=True, use_ff=False, **kwargs):
        super().__init__()
        self.save_hyperparameters()
        self.encoder = RNNEncoder(**kwargs)
        self.use_attn = use_attn
            
        self.dropout = nn.Dropout(self.dropout)
        # self.hidden_size = kwargs["hidden_size"]
        # self.num_layers = kwargs["num_layers"]

        # print(self.use_attn, self.dropout, self.hidden_size)
        if self.use_attn:
            self.dot_attn = DotProductAttention(atten_dim=self.num_layers, use_ff=use_ff)

    def forward(self, emb, lengths=None):
        # M: batch_size x seq_len x nhidden*nlayers
        hidden, M = self.encoder(emb, lengths=lengths)

        layer_outputs = M.split(self.hidden_size, dim=2)
        if self.use_attn:
            output = torch.stack(layer_outputs, dim=2)
            batchs = []
            for i in range(output.size(0)):
                batch_output, _ = self.dot_attn(output[i])
                batchs.append(batch_output)
                
            M = torch.stack(batchs, dim=0)    
            # print("using attn")
            
        else:
            # print("not using attn")
            M = layer_outputs[-1]
        M = M.split(self.hidden_size, dim=2)[-1]
        M = self.dropout(M)
        return  hidden, M
    # def __init__(self,
    #              args,
    #              input_size):
    #     super(Encoder, self).__init__()
    #     self.encoder = RNNEncoder(args.rnn_type,
    #                               input_size,
    #                               args.bidirection,
    #                               args.nlayers,
    #                               args.nhid,
    #                               args.dropout_rnn,
    #                               use_last=False)
    #     self.hidden_size = args.nhid
    #     self.use_all_enc_layers = args.use_all_enc_layers
    #     if self.use_all_enc_layers:
    #         self.layer_weights = nn.Linear(self.hidden_size, 1,
    #                                        bias=False)
    #     self.dropout = nn.Dropout(p=args.dropout_rnn)

    # def count_parameters(self):
    #     return self.encoder.count_parameters()

    # def forward(self, input, input_len):
    #     # M: batch_size x seq_len x nhid*nlayers
    #     hidden, M = self.encoder(input, input_len)
    #     # M: batch_size x seq_len x nhid
    #     layer_outputs = M.split(self.hidden_size, dim=2)
    #     if self.use_all_enc_layers:
    #         output = torch.stack(layer_outputs, dim=2)  # batch_size x seq_len x nlayers x nhid
    #         layer_scores = self.layer_weights(output).squeeze(3)
    #         layer_scores = f.softmax(layer_scores, dim=-1)
    #         M = torch.matmul(output.transpose(2, 3), layer_scores.unsqueeze(3)).squeeze(3)
    #     else:
    #         M = layer_outputs[-1]
    #     M = M.split(self.hidden_size, dim=2)[-1]
    #     M = self.dropout(M)
    #     return hidden, M


if __name__ == "__main__":
    from aicmder.torch import init_seeds
    init_seeds(0)

    rnn_type = "LSTM"
    input_size = 300
    bidirection = True
    num_layers = 6
    hidden_size = 200
    dropout_rnn = 0.2
    encoder = AttnRNNEncoder(rnn_type,
                         bidirectional=bidirection,
                         hidden_size=hidden_size,
                         num_layers=num_layers,
                         dropout=dropout_rnn,
                         input_size=input_size,
                         use_attn=True,
                         use_last=False)

    max_len = 38
    batch_size = 32
    embs = torch.rand(batch_size, max_len, input_size, dtype=torch.float32)
    lengths = [max_len] * batch_size
    print(lengths, type(lengths))
    lengths = torch.Tensor(lengths)
    # lengths = torch.Tensor([max_len] * batch_size, dtype=torch.int64)
    print(lengths.shape, embs.shape)

    encoder_final, memory_bank = encoder(embs, lengths)
    print(len(encoder_final), memory_bank.shape)

    from aicmder.benchmark import model_info

    model_info(encoder, verbose=True)



import cv2
from matplotlib import collections
import numpy as np
import shapely.geometry
import shapely.affinity
from collections import defaultdict
import base64

from fence_detect import RotatedRect, Fence


class Wearing(RotatedRect):

    def set_label(self, label):
        self.label = label

def check_person_wear_helmet(resp_d_person, resp_d, h, w):
    
    obj_rect_set = set()
    for i, obj in enumerate(resp_d["data"]):
        start_x, end_x, end_y, start_y = obj["start_x"], obj["end_x"], obj["end_y"], obj["start_y"]
        width = end_x - start_x
        height = end_y - start_y
        center_x = start_x + width / 2
        center_y = start_y + height / 2
        obj_rect = Wearing()
        obj_rect.set_param(center_x, center_y, width, height, 0)
        obj_rect.set_label(obj["label"])
        obj_rect_set.add(obj_rect)

    resp_d["person"] = []
    for i, obj in enumerate(resp_d_person["data"]):
        start_x, end_x, end_y, start_y = obj["start_x"], obj["end_x"], obj["end_y"], obj["start_y"]
        width = end_x - start_x
        height = end_y - start_y
        center_x = start_x + width / 2
        center_y = start_y + height / 2
        x0, x1, y0, y1 = obj["x0"], obj["x1"], obj["y0"], obj["y1"]
        person_rect = Fence([(x0, y0), (x1, y0), (x1, y1), (x0, y1)], w=w, h=h)
        # person_rect = RotatedRect()
        # person_rect.set_param(center_x, center_y, width, height, 0)
        
        # for debug
        # if obj["label"] == "person 0.69":
        #     print("debug")

        selected_objects = set()
        for obj_rect in obj_rect_set:
            if  "detection" in obj and (("reflective_clothes" in obj["detection"] and "reflective_clothes" in obj_rect.label) or 
                                            ("helmet" in obj["detection"] and "helmet" in obj_rect.label)):
                continue
            if  "detection" in obj and "reflective_clothes" in obj["detection"] and "helmet" in obj["detection"]:
                break
            if person_rect.pointPolygonTest((obj_rect.cx, obj_rect.cy)) == 1:
                if "detection" in obj:
                    obj["detection"] += obj_rect.label + " "
                else:
                    obj["detection"] = obj_rect.label + " "
                selected_objects.add(obj_rect)
        
        obj_rect_set = obj_rect_set.difference(selected_objects)
        resp_d["person"].append(obj)
    
    # print(obj_rect_set)
 



def filter_with_fence(resp_d, fence_list, iouThreshold=0.8):
    new_person = []
    for j, obj in enumerate(resp_d["person"]):
        start_x, end_x, end_y, start_y = obj["start_x"], obj["end_x"], obj["end_y"], obj["start_y"]
        width = end_x - start_x
        height = end_y - start_y
        center_x = start_x + width / 2
        center_y = start_y + height / 2

        obj_rect = RotatedRect()
        obj_rect.set_param(center_x, center_y, width, height, 0)

        for i, fence in enumerate(fence_list):
            if obj_rect.occlusion_rate(fence.fence_rect) >= iouThreshold:
                new_person.append(obj)
                break
    resp_d["person"] = new_person



def draw_restrict_fence(fence_list, img_bgr, resp_d, debug=None):
    if debug is not None and debug > 0:
        try:
            line_thickness = 3
            tl = line_thickness or round(
                0.002 * (img_bgr.shape[0] + img_bgr.shape[1]) / 2) + 1  # line/font thickness
            tf = max(tl - 1, 1)  # font thickness

            for i, fence in enumerate(fence_list):
                # label = str(i)
                cv2.drawContours(img_bgr, fence.contours, 0, (255, 0, 0), 2)
                # print(c1)
            cv2.imwrite('detect_torch_restrict_fence.png', img_bgr)

            # return fence
            if debug > 1:
                with open('detect_torch_restrict_fence.png',  'rb') as img_f:
                    img = img_f.read()
                    resp_d["img"] = base64.b64encode(img).decode('utf8')
        except Exception as e:
            pass


import torch
import torch.distributed as dist
import torch.nn as nn
import time
from torch.nn.parallel import DistributedDataParallel as DDP
import numpy as np
import random
import os
from aicmder.torch.dist_model import check_version as check_version
import inspect
import gc
import collections
try:
    from matplotlib import pyplot as plt
    from matplotlib_inline import backend_inline
    from IPython import display
except:
    pass


def freeze_parameter(model):
    for param in model.parameters():
        param.requires_grad = False


def train_params(model, verbose=False):
    params_to_update = []
    for name, param in model.named_parameters():
        if param.requires_grad == True:
            params_to_update.append(param)
            if verbose:
                print("\t", name)
    return params_to_update


# def set_seed(seed=42):
#     random.seed(seed)
#     os.environ['PYHTONHASHSEED'] = str(seed)
#     np.random.seed(seed)
#     torch.manual_seed(seed)
#     torch.cuda.manual_seed(seed)
#     torch.backends.cudnn.deterministic = True

def init_seeds(seed=0, deterministic=False):
    # Initialize random number generator (RNG) seeds https://pytorch.org/docs/stable/notes/randomness.html
    # cudnn seed 0 settings are slower and more reproducible, else faster and less reproducible
    import torch.backends.cudnn as cudnn

    if deterministic and check_version(torch.__version__, '1.12.0'):  # https://github.com/ultralytics/yolov5/pull/8213
        torch.use_deterministic_algorithms(True)
        os.environ['CUBLAS_WORKSPACE_CONFIG'] = ':4096:8'
        os.environ['PYTHONHASHSEED'] = str(seed)

    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    cudnn.benchmark, cudnn.deterministic = (False, True) if seed == 0 else (True, False)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)  # for Multi-GPU, exception safe


def time_sync():
    # PyTorch-accurate time
    if torch.cuda.is_available():
        torch.cuda.synchronize()
    return time.time()


"""
###################################################################################
add by d2l.ai
###################################################################################
"""


def cpu():
    """Defined in :numref:`sec_use_gpu`"""
    return torch.device('cpu')


def gpu(i=0):
    """Defined in :numref:`sec_use_gpu`"""
    return torch.device(f'cuda:{i}')


def num_gpus():
    """Defined in :numref:`sec_use_gpu`"""
    return torch.cuda.device_count()


def try_gpu(i=0):
    """Return gpu(i) if exists, otherwise return cpu().

    Defined in :numref:`sec_use_gpu`"""
    if num_gpus() >= i + 1:
        return gpu(i)
    return cpu()


def try_all_gpus():
    """Return all available GPUs, or [cpu(),] if no GPU exists.

    Defined in :numref:`sec_use_gpu`"""
    return [gpu(i) for i in range(num_gpus())]


class HyperParameters:
    # def save_hyperparameters(self, ignore=[]):
    #     """Defined in :numref:`sec_oo-design`"""
    #     raise NotImplemented

    def save_hyperparameters(self, ignore=[]):
        """Save function arguments into class attributes.

        Defined in :numref:`sec_utils`"""
        frame = inspect.currentframe().f_back
        _, _, _, local_vars = inspect.getargvalues(frame)
        self.hparams = {k: v for k, v in local_vars.items()
                        if k not in set(ignore+['self']) and not k.startswith('_')}
        for k, v in self.hparams.items():
            if isinstance(v, dict):
                for _k, _v in v.items():
                    try:
                        setattr(self, _k, _v)
                    except:
                        pass
            else:
                setattr(self, k, v)

class Module(nn.Module, HyperParameters):

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.save_hyperparameters()
        self.board = ProgressBoard()

    def to_gpu(self, i=0):
        self.to(gpu(i))

    def to_cpu():
        self.to(cpu())

    def try_gpu(i=0):
        self.to(try_gpu(i))

    def apply_init(self, inputs, init=None):
        """Defined in :numref:`sec_lazy_init`"""
        self.forward(*inputs)
        if init is not None:
            self.net.apply(init)

    def loss(self, y_hat, y):
        raise NotImplementedError

    def forward(self, X):
        raise NotImplementedError
        # assert hasattr(self, 'net'), 'Neural network is defined'
        # return self.net(X)


def use_svg_display():
    """Use the svg format to display a plot in Jupyter.

    Defined in :numref:`sec_calculus`"""
    backend_inline.set_matplotlib_formats('svg')


class ProgressBoard(HyperParameters):
    """Plot data points in animation.

    Defined in :numref:`sec_oo-design`"""

    def __init__(self, xlabel=None, ylabel=None, xlim=None,
                 ylim=None, xscale='linear', yscale='linear',
                 ls=['-', '--', '-.', ':'], colors=['C0', 'C1', 'C2', 'C3'],
                 fig=None, axes=None, figsize=(3.5, 2.5), display=True, is_save=False, save_name="plot.png"):
        self.save_hyperparameters()

    # def draw(self, x, y, label, every_n=1):
    #     raise NotImplemented

    def draw(self, x, y, label, every_n=1):
        """Defined in :numref:`sec_utils`"""
        Point = collections.namedtuple('Point', ['x', 'y'])
        if not hasattr(self, 'raw_points'):
            self.raw_points = collections.OrderedDict()
            self.data = collections.OrderedDict()
        if label not in self.raw_points:
            self.raw_points[label] = []
            self.data[label] = []
        points = self.raw_points[label]
        line = self.data[label]
        points.append(Point(x, y))
        if len(points) != every_n:
            return

        def mean(x): return sum(x) / len(x)
        line.append(Point(mean([p.x for p in points]),
                          mean([p.y for p in points])))
        points.clear()
        if not self.display:
            return
        use_svg_display()
        if self.fig is None:
            self.fig = plt.figure(figsize=self.figsize)
        plt_lines, labels = [], []
        for (k, v), ls, color in zip(self.data.items(), self.ls, self.colors):
            plt_lines.append(plt.plot([p.x for p in v], [p.y for p in v],
                                      linestyle=ls, color=color)[0])
            labels.append(k)
        axes = self.axes if self.axes else plt.gca()
        if self.xlim:
            axes.set_xlim(self.xlim)
        if self.ylim:
            axes.set_ylim(self.ylim)
        if not self.xlabel:
            self.xlabel = self.x
        axes.set_xlabel(self.xlabel)
        axes.set_ylabel(self.ylabel)
        axes.set_xscale(self.xscale)
        axes.set_yscale(self.yscale)
        axes.legend(plt_lines, labels)
        display.display(self.fig)
        if self.is_save:
            plt.savefig(self.save_name)
        display.clear_output(wait=True)


class DataModule(HyperParameters):
    """Defined in :numref:`sec_oo-design`"""

    def __init__(self, *args, **kwargs):
        self.save_hyperparameters()
        super().__init__(*args, **kwargs)

    def get_dataloader(self, train):
        raise NotImplementedError

    def train_dataloader(self):
        return self.get_dataloader(train=True)

    def val_dataloader(self):
        return self.get_dataloader(train=False)

    def get_tensorloader(self, tensors, train, indices=slice(0, None)):
        """Defined in :numref:`sec_synthetic-regression-data`"""
        tensors = tuple(a[indices] for a in tensors)
        dataset = torch.utils.data.TensorDataset(*tensors)
        return torch.utils.data.DataLoader(dataset, self.batch_size,
                                           shuffle=train)


def select_device(device='', batch_size=0):
    # device = 'cpu' or '0' or '0,1,2,3'
    device = str(device).strip().lower().replace(
        'cuda:', '')  # to string, 'cuda:0' to '0'
    cpu = device == 'cpu'
    if cpu:
        # force torch.cuda.is_available() = False
        os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
    elif device:  # non-cpu device requested
        pass
        ######## comment 
        # os.environ['CUDA_VISIBLE_DEVICES'] = device  # set environment variable
        assert torch.cuda.is_available(
        ), f'CUDA unavailable, invalid device {device} requested'  # check availability

    cuda = not cpu and torch.cuda.is_available()
    if cuda:
        # range(torch.cuda.device_count())  # i.e. 0,1,6,7
        devices = device.split(',') if device else ['0']
        n = len(devices)  # device count
        if n > 1 and batch_size > 0:  # check batch_size is divisible by device_count
            assert batch_size % n == 0, f'batch-size {batch_size} not multiple of GPU count {n}'
    return torch.device('cuda:{}'.format(devices[0]) if cuda else 'cpu')


############################### add by faith (modify layer)

def hook_model(model, module_idx=-1, hook_func=None, *args, **kwargs):
    modules_dict = getattr(model, '_modules')
    modules_size = len(modules_dict)
    if module_idx == -1:
        module_idx = modules_size - 1
    for idx, module_name in enumerate(modules_dict):
        if idx == module_idx:
            if hook_func is not None:
                hook_func(modules_dict, module_name, *args, **kwargs)

def modify_last_classifier(modules_dict, module_name, out_features=10):
    '''
        call example: hook_model(model, hook_func=modify_last_classifier, out_features=10)
        # old method
        # last_classifier = list(model.classifier.children())[:-1]
        # model.classifier = torch.nn.Sequential(*last_classifier, torch.nn.Linear(4096, 10))
    '''
    classifier = modules_dict[module_name]
    if type(classifier).__name__ != 'Sequential':
        modules_dict[module_name] = torch.nn.Linear(classifier.in_features, out_features=out_features) 
    else:
        last_classifier = list(classifier.children())[:-1] 
        last_classifier_layer = list(classifier.children())[-1]  
        modules_dict[module_name] = torch.nn.Sequential(*last_classifier, torch.nn.Linear(last_classifier_layer.in_features, out_features=out_features))     


from contextlib import contextmanager
class CudaDecorators(object):
    optimize_cuda_cache = False

    @classmethod
    @contextmanager
    def empty_cuda_cache(cls):
        yield
        if cls.optimize_cuda_cache and torch.cuda.is_available():
            gc.collect()
            torch.cuda.empty_cache()
            gc.collect()


if __name__ == "__main__":
    progress = ProgressBoard(xlabel="x", ylabel="y", is_save=True)
    label = "test"
    x = np.arange(10)
    y = np.arange(10) + 100
    progress.draw(x, y, label)



import cv2
import numpy as np
from fence_detect import RotatedRect


def find_boxes(img):

    # threshold on black color
    lower = np.array([0, 0, 0])
    upper = np.array([100, 100, 100])
    mask = cv2.inRange(img, lower, upper)
    # plt.imshow(mask)
    # apply morphology
    kernel = np.ones((5, 5), np.uint8)
    morph = cv2.morphologyEx(mask, cv2.MORPH_CLOSE, kernel)

    # find lines
    lines = cv2.HoughLinesP(morph, 1, np.pi/180, 100, minLineLength=100, maxLineGap=10)

    # draw lines on copy of input
    result = img.copy()
    # print(lines)
    x_left, x_right = 2000, 0
    for line in lines:
        x1, y1, x2, y2 = line[0]
        if x1 < x_left:
            x_left = x1
        if x2 > x_right:
            x_right = x2
        cv2.line(result, (x1, y1), (x2, y2), (0, 255, 0), 3)
    h, w, _ = img.shape
    print(x_left, x_right, h, w)
    box_w = (x_right - x_left) // 10
    box_h = h
    boxes = []
    for i in range(10):
        boxes.append({
            "start_x": (x_left + i * box_w),
            "end_x": (x_left + (i + 1) * box_w),
            "start_y": 0,
            "end_y": box_h
        })
    # print(boxes)
    return boxes


def match_ans_with_boxes(mc_data, boxes_data, iouThreshold=0.8):
    boxes = []
    for obj in boxes_data:
        start_x, end_x, end_y, start_y = obj["start_x"], obj["end_x"], obj["end_y"], obj["start_y"]
        width = end_x - start_x
        height = end_y - start_y
        center_x = start_x + width / 2
        center_y = start_y + height / 2

        obj_rect = RotatedRect()
        obj_rect.set_param(center_x, center_y, width, height, 0)
        boxes.append(obj_rect)

    ans = []
    for j, obj in enumerate(mc_data):
        start_x, end_x, end_y, start_y = obj["start_x"], obj["end_x"], obj["end_y"], obj["start_y"]
        width = end_x - start_x
        height = end_y - start_y
        center_x = start_x + width / 2
        center_y = start_y + height / 2

        obj_rect = RotatedRect()
        obj_rect.set_param(center_x, center_y, width, height, 0)

        for i, box in enumerate(boxes):
            if obj_rect.occlusion_rate(box) >= iouThreshold:
                ans.append(((i + 1), obj))
                break
    return ans



# src: https://github.com/facebookresearch/DrQA/blob/master/drqa/reader/data.py
import unicodedata
import numpy as np
import random
PAD = 0
UNK = 1
BOS = 2
EOS = 3

PAD_WORD = '<blank>'
UNK_WORD = '<unk>'
BOS_WORD = '<s>'
EOS_WORD = '</s>'


class Vocabulary(object):
    def __init__(self, no_special_token=False):
        if no_special_token:
            self.tok2ind = {PAD_WORD: PAD,
                            UNK_WORD: UNK}
            self.ind2tok = {PAD: PAD_WORD,
                            UNK: UNK_WORD}
        else:
            self.tok2ind = {PAD_WORD: PAD,
                            UNK_WORD: UNK,
                            BOS_WORD: BOS,
                            EOS_WORD: EOS}
            self.ind2tok = {PAD: PAD_WORD,
                            UNK: UNK_WORD,
                            BOS: BOS_WORD,
                            EOS: EOS_WORD}

    @staticmethod
    def normalize(token):
        return unicodedata.normalize('NFD', token)

    def __len__(self):
        return len(self.tok2ind)

    def __iter__(self):
        return iter(self.tok2ind)

    def __contains__(self, key):
        if type(key) == int:
            return key in self.ind2tok
        elif type(key) == str:
            return self.normalize(key) in self.tok2ind

    def __getitem__(self, key):
        if type(key) == int:
            return self.ind2tok.get(key, UNK_WORD)
        elif type(key) == str:
            return self.tok2ind.get(self.normalize(key),
                                    self.tok2ind.get(UNK_WORD))
        else:
            raise RuntimeError('Invalid key type.')

    def __setitem__(self, key, item):
        if type(key) == int and type(item) == str:
            self.ind2tok[key] = item
        elif type(key) == str and type(item) == int:
            self.tok2ind[key] = item
        else:
            raise RuntimeError('Invalid (key, item) types.')

    def add(self, token):
        token = self.normalize(token)
        if token not in self.tok2ind:
            index = len(self.tok2ind)
            self.tok2ind[token] = index
            self.ind2tok[index] = token

    def add_tokens(self, token_list):
        assert isinstance(token_list, list)
        for token in token_list:
            self.add(token)

    def tokens(self):
        """Get dictionary tokens.
        Return all the words indexed by this dictionary, except for special
        tokens.
        """
        tokens = [k for k in self.tok2ind.keys()
                  if k not in {PAD_WORD, UNK_WORD}]
        return tokens

    def remove(self, key):
        if key in self.tok2ind:
            ind = self.tok2ind[key]
            del self.tok2ind[key]
            del self.ind2tok[ind]
            return True
        return False


class UnicodeCharsVocabulary(Vocabulary):
    """Vocabulary containing character-level and word level information.
    Has a word vocabulary that is used to lookup word ids and
    a character id that is used to map words to arrays of character ids.
    The character ids are defined by ord(c) for c in word.encode('utf-8')
    This limits the total number of possible char ids to 256.
    To this we add 5 additional special ids: begin sentence, end sentence,
        begin word, end word and padding.
    """

    def __init__(self, words, max_word_length,
                 no_special_token):
        super(UnicodeCharsVocabulary, self).__init__(no_special_token)
        self._max_word_length = max_word_length

        # char ids 0-255 come from utf-8 encoding bytes
        # assign 256-300 to special chars
        self.bow_char = 256  # <begin word>
        self.eow_char = 257  # <end word>
        self.pad_char = 258  # <padding>

        for w in words:
            self.add(w)
        num_words = len(self.ind2tok)

        self._word_char_ids = np.zeros([num_words, max_word_length],
                                       dtype=np.int32)

        for i, word in self.ind2tok.items():
            self._word_char_ids[i] = self._convert_word_to_char_ids(word)

    @property
    def word_char_ids(self):
        return self._word_char_ids

    @property
    def max_word_length(self):
        return self._max_word_length

    def _convert_word_to_char_ids(self, word):
        code = np.zeros([self.max_word_length], dtype=np.int32)
        code[:] = self.pad_char

        word_encoded = word.encode('utf-8', 'ignore')[:(self.max_word_length - 2)]
        code[0] = self.bow_char
        for k, chr_id in enumerate(word_encoded, start=1):
            code[k] = chr_id
        code[k + 1] = self.eow_char

        return code

    def word_to_char_ids(self, word):
        if word in self.tok2ind:
            return self._word_char_ids[self.tok2ind[word]]
        else:
            return self._convert_word_to_char_ids(word)

    def encode_chars(self, sentence, split=True):
        '''
        Encode the sentence as a white space delimited string of tokens.
        '''
        if split:
            chars_ids = [self.word_to_char_ids(cur_word)
                         for cur_word in sentence.split()]
        else:
            chars_ids = [self.word_to_char_ids(cur_word)
                         for cur_word in sentence]

        return chars_ids


class RandomGenerator:
    """Randomly draw among {1, ..., n} according to n sampling weights."""

    def __init__(self, sampling_weights):
        """Defined in :numref:`sec_word2vec_data`"""
        # Exclude
        self.population = list(range(1, len(sampling_weights) + 1))
        self.sampling_weights = sampling_weights
        self.candidates = []
        self.i = 0

    def draw(self):
        if self.i == len(self.candidates):
            # Cache `k` random sampling results
            self.candidates = random.choices(
                self.population, self.sampling_weights, k=10000)
            self.i = 0
        self.i += 1
        return self.candidates[self.i - 1]


def get_negatives(all_contexts, vocab, counter, K):
    """Return noise words in negative sampling.

    Defined in :numref:`sec_word2vec_data`"""
    # Sampling weights for words with indices 1, 2, ... (index 0 is the
    # excluded unknown token) in the vocabulary
    sampling_weights = [counter[vocab.to_tokens(i)]**0.75
                        for i in range(1, len(vocab))]
    all_negatives, generator = [], RandomGenerator(sampling_weights)
    for contexts in all_contexts:
        negatives = []
        while len(negatives) < len(contexts) * K:
            neg = generator.draw()
            # Noise words cannot be context words
            if neg not in contexts:
                negatives.append(neg)
        all_negatives.append(negatives)
    return all_negatives


reshape = lambda x, *args, **kwargs: x.reshape(*args, **kwargs)


def batchify(data):
    """Return a minibatch of examples for skip-gram with negative sampling.

    Defined in :numref:`sec_word2vec_data`"""
    max_len = max(len(c) + len(n) for _, c, n in data)
    centers, contexts_negatives, masks, labels = [], [], [], []
    for center, context, negative in data:
        cur_len = len(context) + len(negative)
        centers += [center]
        contexts_negatives += [context + negative + [0] * (max_len - cur_len)]
        masks += [[1] * cur_len + [0] * (max_len - cur_len)]
        labels += [[1] * len(context) + [0] * (max_len - len(context))]
    return (reshape(torch.tensor(centers), (-1, 1)), torch.tensor(
        contexts_negatives), torch.tensor(masks), torch.tensor(labels))


if __name__ == "__main__":
    words = set()
    words.add("hello")
    words.add("world")

    dictioanry = UnicodeCharsVocabulary(words, 30, True)
    print(dictioanry, len(dictioanry))



import cv2
import numpy as np
import time

def plot_one_box(x, im, color=(128, 128, 128), label=None, line_thickness=1):
    # Plots one bounding box on image 'im' using OpenCV
    assert im.data.contiguous, 'Image not contiguous. Apply np.ascontiguousarray(im) to plot_on_box() input image.'
    tl = line_thickness or round(0.002 * (im.shape[0] + im.shape[1]) / 2) + 1  # line/font thickness
    c1, c2 = (int(x[0]), int(x[1])), (int(x[2]), int(x[3]))
    cv2.rectangle(im, c1, c2, color, thickness=tl, lineType=cv2.LINE_AA)
    if label:
        tf = max(tl - 1, 1)  # font thickness
        t_size = cv2.getTextSize(label, 0, fontScale=tl / 3, thickness=tf)[0]
        c2 = c1[0] + t_size[0], c1[1] - t_size[1] - 3
        cv2.rectangle(im, c1, c2, color, -1, cv2.LINE_AA)  # filled
        cv2.putText(im, label, (c1[0], c1[1] - 2), 0, tl / 3, [225, 255, 255], thickness=tf, lineType=cv2.LINE_AA)


def letterbox(im, new_shape=(640, 640), color=(114, 114, 114), auto=True, scaleFill=False, scaleup=True, stride=32):
    # Resize and pad image while meeting stride-multiple constraints
    shape = im.shape[:2]  # current shape [height, width]
    if isinstance(new_shape, int):
        new_shape = (new_shape, new_shape)

    # Scale ratio (new / old)
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    if not scaleup:  # only scale down, do not scale up (for better val mAP)
        r = min(r, 1.0)

    # Compute padding
    ratio = r, r  # width, height ratios
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
    if auto:  # minimum rectangle
        dw, dh = np.mod(dw, stride), np.mod(dh, stride)  # wh padding
    elif scaleFill:  # stretch
        dw, dh = 0.0, 0.0
        new_unpad = (new_shape[1], new_shape[0])
        ratio = new_shape[1] / shape[1], new_shape[0] / shape[0]  # width, height ratios

    dw /= 2  # divide padding into 2 sides
    dh /= 2

    if shape[::-1] != new_unpad:  # resize
        im = cv2.resize(im, new_unpad, interpolation=cv2.INTER_LINEAR)

    # tensor = numpy_helper.from_array(im)
    # with open('resize.pb', 'wb') as f:
    #     f.write(tensor.SerializeToString())


    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    im = cv2.copyMakeBorder(im, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)  # add border
    return im, ratio, (dw, dh)


def xywh2xyxy(x):
    # Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
    y = np.copy(x)
    y[:, 0] = x[:, 0] - x[:, 2] / 2  # top left x
    y[:, 1] = x[:, 1] - x[:, 3] / 2  # top left y
    y[:, 2] = x[:, 0] + x[:, 2] / 2  # bottom right x
    y[:, 3] = x[:, 1] + x[:, 3] / 2  # bottom right y
    return y



def nms_cpu(boxes, scores, overlap_threshold=0.5, min_mode=False):
    # boxes = boxes.cpu().numpy()
    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]
    # scores = boxes[:, 4]

    areas = (x2 - x1 + 1) * (y2 - y1 + 1)
    order = scores.argsort()[::-1]

    keep = []
    while order.size > 0:
        keep.append(order[0])
        xx1 = np.maximum(x1[order[0]], x1[order[1:]])
        yy1 = np.maximum(y1[order[0]], y1[order[1:]])
        xx2 = np.minimum(x2[order[0]], x2[order[1:]])
        yy2 = np.minimum(y2[order[0]], y2[order[1:]])

        w = np.maximum(0.0, xx2 - xx1 + 1)
        h = np.maximum(0.0, yy2 - yy1 + 1)
        inter = w * h

        if min_mode:
            ovr = inter / np.minimum(areas[order[0]], areas[order[1:]])
        else:
            ovr = inter / (areas[order[0]] + areas[order[1:]] - inter)

        inds = np.where(ovr <= overlap_threshold)[0]
        order = order[inds + 1]
    return keep


def non_max_suppression(prediction, conf_thres=0.25, iou_thres=0.45, classes=None, agnostic=False, multi_label=False,
                        labels=(), max_det=300):
    """Runs Non-Maximum Suppression (NMS) on inference results

    Returns:
         list of detections, on (n,6) tensor per image [xyxy, conf, cls]
    """
    print("prediction", prediction.shape)
    nc = prediction.shape[2] - 5  # number of classes
    xc = prediction[..., 4] > conf_thres  # candidates

    # Checks
    assert 0 <= conf_thres <= 1, f'Invalid Confidence threshold {conf_thres}, valid values are between 0.0 and 1.0'
    assert 0 <= iou_thres <= 1, f'Invalid IoU {iou_thres}, valid values are between 0.0 and 1.0'

    # Settings
    min_wh, max_wh = 2, 4096  # (pixels) minimum and maximum box width and height
    max_nms = 30000  # maximum number of boxes into torchvision.ops.nms()
    time_limit = 10.0  # seconds to quit after
    redundant = True  # require redundant detections
    multi_label &= nc > 1  # multiple labels per box (adds 0.5ms/img)
    merge = False  # use merge-NMS

    t = time.time()
    # output = [torch.zeros((0, 6), device=prediction.device)] * prediction.shape[0]
    output = []
    for xi, x in enumerate(prediction):  # image index, image inference
        # Apply constraints
        # x[((x[..., 2:4] < min_wh) | (x[..., 2:4] > max_wh)).any(1), 4] = 0  # width-height
        x = x[xc[xi]]  # confidence

        # Cat apriori labels if autolabelling
        if labels and len(labels[xi]):
            l = labels[xi]
            v = np.zeros((len(l), nc + 5), device=x.device)
            v[:, :4] = l[:, 1:5]  # box
            v[:, 4] = 1.0  # conf
            v[range(len(l)), l[:, 0].long() + 5] = 1.0  # cls
            x = np.cat((x, v), 0)

        # If none remain process next image
        if not x.shape[0]:
            continue

        # Compute conf
        x[:, 5:] *= x[:, 4:5]  # conf = obj_conf * cls_conf

        # Box (center x, center y, width, height) to (x1, y1, x2, y2)
        box = xywh2xyxy(x[:, :4])

        # Detections matrix nx6 (xyxy, conf, cls)
        if multi_label:
            i, j = (x[:, 5:] > conf_thres).nonzero(as_tuple=False).T
            x = np.cat((box[i], x[i, j + 5, None], j[:, None].float()), 1)
        else:  # best class only
            conf = x[:, 5:].max(1, keepdims=True)
            x = np.concatenate((box, conf), 1)[np.squeeze(conf) > conf_thres]

        # Filter by class  Only one class
        # if classes is not None:
        #     x = x[(x[:, 5:6] == torch.tensor(classes, device=x.device)).any(1)]

        # Apply finite constraint
        # if not torch.isfinite(x).all():
        #     x = x[torch.isfinite(x).all(1)]

        # Check shape
        n = x.shape[0]  # number of boxes
        if not n:  # no boxes
            continue
        elif n > max_nms:  # excess boxes
            x = x[x[:, 4].argsort(descending=True)[:max_nms]]  # sort by confidence

        # Batched NMS

        # only one class
        # c = x[:, 5:6] * (0 if agnostic else max_wh)  # classes
        boxes, scores = x[:, :4], x[:, 4]  # boxes (offset by class), scores
        i = nms_cpu(boxes, scores, iou_thres)  # NMS
        if len(i) > max_det:  # limit detections
            i = i[:max_det]


        output.append(x[i])
        if (time.time() - t) > time_limit:
            print(f'WARNING: NMS time limit {time_limit}s exceeded')
            break  # time limit exceeded

    return output


def scale_coords(img1_shape, coords, img0_shape, ratio_pad=None):
    # Rescale coords (xyxy) from img1_shape to img0_shape
    if ratio_pad is None:  # calculate from img0_shape
        gain = min(img1_shape[0] / img0_shape[0], img1_shape[1] / img0_shape[1])  # gain  = old / new
        pad = (img1_shape[1] - img0_shape[1] * gain) / 2, (img1_shape[0] - img0_shape[0] * gain) / 2  # wh padding
    else:
        gain = ratio_pad[0][0]
        pad = ratio_pad[1]

    coords[:, [0, 2]] -= pad[0]  # x padding
    coords[:, [1, 3]] -= pad[1]  # y padding
    coords[:, :4] /= gain
    clip_coords(coords, img0_shape)
    return coords

def clip_coords(boxes, shape):
    # Clip bounding xyxy bounding boxes to image shape (height, width)
    boxes[:, [0, 2]] = boxes[:, [0, 2]].clip(0, shape[1])  # x1, x2
    boxes[:, [1, 3]] = boxes[:, [1, 3]].clip(0, shape[0])  # y1, y2



def box_area(boxes):
    return (boxes[:, 2] - boxes[:, 0]) * (boxes[:, 3] - boxes[:, 1])


def box_iou(box1, box2):
    area1 = box_area(box1)  # N
    area2 = box_area(box2)  # M
    # broadcasting, 两个数组各维度大小 从后往前对比一致， 或者 有一维度值为1；
    lt = np.maximum(box1[:, np.newaxis, :2], box2[:, :2])
    rb = np.minimum(box1[:, np.newaxis, 2:], box2[:, 2:])
    wh = rb - lt
    wh = np.maximum(0, wh) # [N, M, 2]
    inter = wh[:, :, 0] * wh[:, :, 1]
    iou = inter / (area1[:, np.newaxis] + area2 - inter)
    return iou  # NxM


def numpy_nms(boxes, scores, iou_threshold :float):
    idxs = scores.argsort()  # 按分数 降序排列的索引 [N]
    keep = []
    while idxs.size > 0:  # 统计数组中元素的个数
        max_score_index = idxs[-1]
        max_score_box = boxes[max_score_index][None, :]
        keep.append(max_score_index)
        if idxs.size == 1:
            break
        idxs = idxs[:-1]  # 将得分最大框 从索引中删除； 剩余索引对应的框 和 得分最大框 计算IoU；
        other_boxes = boxes[idxs]  # [?, 4]
        ious = box_iou(max_score_box, other_boxes)  # 一个框和其余框比较 1XM
        idxs = idxs[ious[0] <= iou_threshold]
    keep = np.array(keep)  
    return keep


def bbox_iou(box1, box2, x1y1x2y2=True):
    """
    description: compute the IoU of two bounding boxes
    param:
        box1: A box coordinate (can be (x1, y1, x2, y2) or (x, y, w, h))
        box2: A box coordinate (can be (x1, y1, x2, y2) or (x, y, w, h))
        x1y1x2y2: select the coordinate format
    return:
        iou: computed iou
    """
    if not x1y1x2y2:
        # Transform from center and width to exact coordinates
        b1_x1, b1_x2 = box1[:, 0] - box1[:, 2] / 2, box1[:, 0] + box1[:, 2] / 2
        b1_y1, b1_y2 = box1[:, 1] - box1[:, 3] / 2, box1[:, 1] + box1[:, 3] / 2
        b2_x1, b2_x2 = box2[:, 0] - box2[:, 2] / 2, box2[:, 0] + box2[:, 2] / 2
        b2_y1, b2_y2 = box2[:, 1] - box2[:, 3] / 2, box2[:, 1] + box2[:, 3] / 2
    else:
        # Get the coordinates of bounding boxes
        b1_x1, b1_y1, b1_x2, b1_y2 = box1[:, 0], box1[:, 1], box1[:, 2], box1[:, 3]
        b2_x1, b2_y1, b2_x2, b2_y2 = box2[:, 0], box2[:, 1], box2[:, 2], box2[:, 3]

    # Get the coordinates of the intersection rectangle
    inter_rect_x1 = np.maximum(b1_x1, b2_x1)
    inter_rect_y1 = np.maximum(b1_y1, b2_y1)
    inter_rect_x2 = np.minimum(b1_x2, b2_x2)
    inter_rect_y2 = np.minimum(b1_y2, b2_y2)
    # Intersection area
    inter_area = np.clip(inter_rect_x2 - inter_rect_x1 + 1, 0, None) * \
                    np.clip(inter_rect_y2 - inter_rect_y1 + 1, 0, None)
    # Union Area
    b1_area = (b1_x2 - b1_x1 + 1) * (b1_y2 - b1_y1 + 1)
    b2_area = (b2_x2 - b2_x1 + 1) * (b2_y2 - b2_y1 + 1)

    iou = inter_area / (b1_area + b2_area - inter_area + 1e-16)

    return iou


from enum import unique
from aicmder.torch.dataset import InfiniteDataLoader
from huggingface_hub import hf_hub_url, cached_download
import os
# os.environ["CUDA_VISIBLE_DEVICES"] = "0"
os.environ["OMP_NUM_THREADS"] = "16"
from tqdm import tqdm
from PIL import Image
import numpy as np
from pathlib import Path

import torch
import torchvision
from torchvision import datasets, transforms, models
from torch.utils.data import Dataset
from torch.utils.data import BatchSampler, DataLoader
from torchvision import models
import torch.nn as nn
from fp16module import FP16Module
import time
from torch.optim import lr_scheduler
from copy import deepcopy
from datetime import datetime
import glob
import base64
import random
from torchmetrics.functional import precision_recall
from PIL import ImageFile
ImageFile.LOAD_TRUNCATED_IMAGES = True
from aicmder.torch import freeze_parameter, train_params
from CustomNet import FC
from aicmder.torch.dist_model import *
from torch.utils.data import DataLoader, Dataset, distributed
from aicmder.torch.dist_model import Trainable, load_model, smart_resume
from aicmder.utils import Callbacks, methods
from aicmder.torch.sampler import DistributedSamplerWrapper


def evaluate_func(fitness, best_fitness):
    best_train_acc, best_train_loss = best_fitness
    train_acc, train_loss = fitness
    if train_acc > best_train_acc or (train_acc == best_train_acc and best_train_loss is not None and train_loss < best_train_loss):
        return True
    else: 
        return False
    
def read_image_rgb(path):
    pil_img = Image.open(path)
    pil_img.load()
    if pil_img.format is 'PNG' and pil_img.mode is not 'RGBA':
        pil_img = pil_img.convert('RGBA')
    pil_img = pil_img.convert('RGB')
    return pil_img


def vprint(*args, verbose=True):
    if verbose:
        print(*args)


def seed_worker(worker_id):
    # Set dataloader worker seed https://pytorch.org/docs/stable/notes/randomness.html#dataloader
    worker_seed = torch.initial_seed() % 2 ** 32
    np.random.seed(worker_seed)
    random.seed(worker_seed)


classifier_transforms = transforms.Compose([
    transforms.Resize((320, 320)),
    transforms.ToTensor(),
    transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
])


# def predict_image(pil_image, model, device):
#     input_img = classifier_transforms(pil_image).float().unsqueeze(0)
#     outputs = model(input_img.to(device))
#     result = torch.max(outputs, 1)[1].cpu().reshape(-1).tolist()[0]
#     return result


class ImageDataset(Dataset):

    def __init__(self, objects):
        self.objects = objects
        self.resnet_transforms = classifier_transforms

    def __len__(self):
        return len(self.objects)

    def __getitem__(self, idx):
        obj = self.objects[idx]
        assert isinstance(obj, (str, np.ndarray, Image.Image))
        if isinstance(obj, str):
            pil_img = read_image_rgb(obj)
        elif isinstance(obj, np.ndarray):
            pil_img = Image.fromarray(obj)
        elif isinstance(obj, Image.Image):
            pil_img = obj

        resnet_img = self.resnet_transforms(pil_img).float()

        return resnet_img





class WatermarksPredictor:

    def __init__(self, model, device, feature_extract=True, workers=os.cpu_count(), bs=8, verbose=False, save_dir="./", name='resnext101_32x8d-large', fp16=False, cache_dir='weights/'):
        """
        Predict with watermark classifier using batches and torch.DataLoader
        resnet_model: watermark classifier model
        device: torch.device to use
        workers: number of workers for dataloader
        bs: batch size to use
        verbose: print additional info or not
        """
        config = MODELS[name]
        self.model = config['resnet'](pretrained=True)
        # pretrain_statedict = self.model.state_dict()

        if feature_extract:
            # Freeze
            freeze_parameter(self.model)
            # for param in self.model.parameters():
            #     param.requires_grad = False
        num_ftrs = self.model.fc.in_features

        # self.model.fc = nn.Sequential(nn.Linear(num_ftrs, 512),
        #                                 nn.ReLU(),
        #                                 nn.Dropout(0.2),
        #                                 nn.Linear(512, 2),
        #                                 nn.LogSoftmax(dim=1))
        self.model.fc = FC(num_ftrs)
        # self.model.fc = nn.Linear(num_ftrs, 2)

        if isinstance(model, str):
            # path, load checkpoint first
            # load checkpoint to CPU to avoid CUDA memory leak
            # import os, sys
            # sys_path = os.path.join(os.path.dirname(__file__), "CustomNet.py")
            # sys.path.append(sys_path)
            # print(sys_path)

            # ckpt = torch.load(model, map_location='cpu')
            # # csd = ckpt['model'].float().state_dict()
            # csd = ckpt['model']
            # self.model.load_state_dict(csd, strict=False)  # load
            
            self.model, self.ckpt = load_model(model, is_val=False)
            
        else:
            config_file_url = hf_hub_url(
                repo_id=config['repo_id'], filename=config['filename'])
            cached_download(config_file_url, cache_dir=cache_dir,
                            force_filename=config['filename'])
            weights = torch.load(os.path.join(cache_dir, config['filename']), device)

            # for pretrain, train in zip(pretrain_statedict, weights):
            #     print(pretrain, train)
            # self.model.load_state_dict(weights)

        if fp16:
            self.model = FP16Module(self.model)

        self.model = self.model.to(device)
        self.model.eval()

        # model_info(self.model)        
        
        workers = 8 
        self.num_workers = workers
        self.bs = bs
        self.device = device
        self.verbose = verbose

        vprint(f'Using device {self.device}', verbose=self.verbose)
        save_dir = Path(save_dir)
        w = save_dir / 'train'  # weights dir
        w.mkdir(parents=True, exist_ok=True)  # make dir
        self.last, self.best = w / 'last.pt', w / 'best.pt'
        self.input_size = 320

    @smart_inference_mode()
    def run(self, files):
        """
        Processes input objects (list of paths, list of PIL images, list of numpy arrays) and returns model results.
        files: objects to process. Should be list of paths to images or list of PIL images or list of numpy arrays
        """
        # vprint(f'Files to process: {len(files)}', verbose=self.verbose)
        if isinstance(files, list):
            eval_dataset = ImageDataset(files)
        else:
            eval_dataset = datasets.ImageFolder(files, transforms.Compose([
                transforms.Resize((self.input_size, self.input_size)),
                transforms.ToTensor(),
                transforms.Normalize([0.485, 0.456, 0.406], [
                                     0.229, 0.224, 0.225])
            ]))

        loader = DataLoader(
            eval_dataset,
            sampler=torch.utils.data.SequentialSampler(eval_dataset),
            batch_size=self.bs,
            drop_last=False,
            num_workers=self.num_workers
        )
        if self.verbose:
            loader = tqdm(loader)

        preds, targets = [], []
        if isinstance(files, list):
            for batch in loader:
                with torch.no_grad():
                    outputs = self.model(batch.to(self.device))
                    _, pred = torch.max(outputs, 1)
                    preds.extend(pred.cpu().reshape(-1).tolist())

        else:
            for batch, target in loader:
                try:
                    with torch.no_grad():
                        outputs = self.model(batch.to(self.device))
                        _, pred = torch.max(outputs, 1)
                        preds.extend(pred.cpu().reshape(-1).tolist())
                        targets.extend(target.cpu().reshape(-1).tolist())
                except Exception as e:
                    pass

            preds = torch.tensor(preds)
            targets = torch.tensor(targets)
            # print(preds.shape, targets.shape)
            print(precision_recall(preds, targets, average='macro', num_classes=2))
        return preds

    # def freeze_conv_param(self):
    #     for v in self.model.modules():
    #         if isinstance(v, nn.modules.conv.Conv2d) and hasattr(v, 'weight') and isinstance(v.weight, nn.Parameter):
    #             # print("--")
    #             v.weight.requires_grad = False
    #         elif isinstance(v, torchvision.models.resnet.Bottleneck):
    #             if hasattr(v, 'conv1'):
    #                 v.conv1.weight.requires_grad = False
    #             if hasattr(v, 'conv2'):
    #                 v.conv2.weight.requires_grad = False
    #             if hasattr(v, 'conv3'):
    #                 v.conv3.weight.requires_grad = False
    #         else:
    #             print(type(v))

    def train(self, traindir, criterion=nn.CrossEntropyLoss(), start_epoch=0, epochs=1, opt='SGD', lr=0.01, lrf=0.01, momentum=0.937, weight_decay=0.0005, amp=False):
        # vprint(f'Files to process: {len(files)}', verbose=self.verbose)
        # train_dataset = ImageDataset(files)
        # loader = DataLoader(
        #     train_dataset,
        #     sampler=None, #torch.utils.data.SequentialSampler(train_dataset),
        #     batch_size=self.bs,
        #     pin_memory=True,
        #     drop_last=False,
        #     num_workers=self.num_workers
        # )

        trainable = Trainable(self.bs)

        # if isinstance(self.model, torch.nn.Sequential):
        #     functions = list(self.model.children())

        # for i, (name, layer) in enumerate(self.model.named_modules()):
        #     if isinstance(layer, nn.ReLU):
        #         print(layer)

        trainable.on_pretrain_routine_start(self.model, verbose=True)
        # callbacks = Callbacks()
        # for k in methods(trainable):
        #     callbacks.register_action(k, callback=getattr(trainable, k))

        self.model, self.device, amp, scaler, ema = trainable.on_train_start(model=self.model, device=self.device, evaluate_func=evaluate_func, patience=5)
       
        ################################ distributed LOCAL_RANK 
        with torch_distributed_zero_first(LOCAL_RANK):
            image_dataset = datasets.ImageFolder(traindir, transforms.Compose([
                transforms.RandomResizedCrop(self.input_size),
                transforms.RandomHorizontalFlip(),
                # transforms.Resize((320, 320)),
                transforms.ToTensor(),
                transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
            ]))

        np_target = np.array(image_dataset.targets)
        unique_target = np.unique(np_target)

        # https://discuss.pytorch.org/t/how-to-handle-imbalanced-classes/11264/2
        class_sample_count = np.array(
            [len(np.where(np_target == t)[0]) for t in unique_target])
        weight = 1. / class_sample_count
        samples_weight = np.array([weight[t] for t in np_target])
        samples_weight = torch.from_numpy(samples_weight)
        samples_weight = samples_weight.double()
        sampler = torch.utils.data.WeightedRandomSampler(
            samples_weight, len(samples_weight))

        if LOCAL_RANK != -1:
            # https://github.com/Lightning-AI/lightning/issues/9883
            if torch.distributed.is_initialized():
                print("running DistributedSamplerWrapper!")
                sampler = DistributedSamplerWrapper(sampler,
                                            shuffle=True,
                                            num_replicas=WORLD_SIZE,
                                            rank=LOCAL_RANK)
            else:
                sampler = distributed.DistributedSampler(image_dataset, shuffle=False)
        
        
        # target = torch.from_numpy(target).long()
        # train_dataset = torch.utils.data.TensorDataset(data, target)

        # print("size", train_size, image_dataset.imgs, samples_weight)
        print("train num works", self.num_workers)
        loader = InfiniteDataLoader(
            image_dataset,
            # torch.utils.data.SequentialSampler(train_dataset),
            sampler=sampler,
            shuffle=False,
            batch_size=self.bs // WORLD_SIZE,
            pin_memory=True,
            drop_last=False,
            worker_init_fn=seed_worker,
            num_workers=self.num_workers
        )
        train_size = len(loader.dataset)  # len(image_dataset.imgs)

        # nb = len(loader)  # number of batches
        criterion = criterion
        optimizer = smart_optimizer(self.model, opt, lr, momentum, weight_decay)

        # params_to_update = self.model.parameters()
        # params_to_update = train_params(de_parallel(self.model))
        # optimizer = torch.optim.SGD(params_to_update, lr=0.001, momentum=0.9)

        def lr_lambda(x): return (1 - x / epochs) * (1.0 - lrf) + lrf  # linear
        scheduler = lr_scheduler.LambdaLR(optimizer, lr_lambda=lr_lambda)

        # self.freeze_conv_param()
        best_train_acc = 0.0
        best_train_loss = None

        ############################ resume now
        if hasattr(self, 'ckpt'):
            best_fitness, start_epoch, epochs = smart_resume(self.ckpt, optimizer=optimizer, ema=ema, epochs=epochs, resume=True)
            best_train_acc, best_train_loss = best_fitness 
            scheduler.last_epoch = start_epoch - 1  # do not move
        ############################ resume now
        for epoch in range(start_epoch, epochs):
            train_count = 0
            
            pbar, train_size = trainable.on_train_epoch_start(self.model, epoch, loader)
            
            running_loss = 0.0
            running_corrects = 0

            for batch_idx, (inputs, labels) in pbar:
                # number of iteration
                ni = batch_idx + epoch * train_size

                # print(batch_idx, inputs.shape, labels)
                # inputs = torch.autograd.Variable(inputs)
                inputs = inputs.to(self.device)
                labels = labels.to(self.device)

                # optimizer.zero_grad()

                with torch.cuda.amp.autocast(amp):
                    outputs = self.model(inputs)
                    _, preds = torch.max(outputs, 1)

                    # Input(N, C) where C = number of classes
                    # target(N) where each value is 0 <= targets[i] <= C - 1
                    loss = criterion(outputs, labels)
                    if RANK != -1:
                        loss *= WORLD_SIZE  # gradient averaged between devices in DDP mode
               

                if RANK in {-1, 0}:
                    running_loss += loss.item() * inputs.size(0)
                    running_corrects += torch.sum(preds == labels.data)
                    train_count += inputs.size(0)
                    
                try:
                    mem = f'{torch.cuda.memory_reserved() / 1E9 if torch.cuda.is_available() else 0:.3g}G'  # (GB)
                    desc = ('%15s' * 2 + '%15s' * 4) % (f'{epoch}/{epochs - 1}', f'Mem:{mem}', "Loss: {:.2f}".format (running_loss), "Correct: {}".format(running_corrects), "Cnt: {}".format(train_count), "Acc: {:2f}".format(running_corrects / train_count if train_count != 0 else 0))      
                except Exception as e:
                    desc = ''          
                # zero gradient or use accumulate gradient 
                trainable.on_train_batch_end(self.model, scaler, optimizer, loss, desc=desc, ni=ni)

            # end batch
            lr = [x['lr'] for x in optimizer.param_groups]  # for loggers
            

            if RANK in {-1, 0}:
                epoch_loss = running_loss / train_count
                epoch_acc = running_corrects.double() / train_count
                
                print('training {} Loss: {:.4f} Acc: {:.4f} Count: {} Train_size: {}'.format(
                    epoch, epoch_loss, epoch_acc, train_count, train_size))

                if epoch_acc > best_train_acc or (epoch_acc == best_train_acc and best_train_loss is not None and epoch_loss < best_train_loss):
                    best_train_acc = epoch_acc
                    best_train_loss = epoch_loss
                    # state_dict = self.model.state_dict()
                    # ckpt = {
                    #     'epoch': epoch,
                    #     # 'model': deepcopy(self.model),
                    #     'model': state_dict,
                    #     'optimizer': optimizer.state_dict(),
                    #     # 'wandb_id': None,  # loggers.wandb.wandb_run.id if loggers.wandb else None,
                    #     'date': datetime.now().isoformat()}
                    # torch.save(ckpt, self.best)
                    # del ckpt
                    trainable.on_model_save(self.model, epoch, optimizer, (best_train_acc, best_train_loss), self.best) 
            
            stop = trainable.on_train_epoch_end(epoch, (best_train_acc, best_train_loss), scheduler=scheduler)
            if stop:
                break
                    
        trainable.on_model_save(self.model, epoch, optimizer, (best_train_acc, best_train_loss), self.last)
        trainable.on_train_end()
    


MODELS = {
    'resnext101_32x8d-large': dict(
        resnet=models.resnext101_32x8d,
        repo_id='boomb0om/dataset-filters',
        filename='watermark_classifier-resnext101_32x8d-input_size320-4epochs_c097_w082.pth',
    ),
    'resnext50_32x4d-small': dict(
        resnet=models.resnext50_32x4d,
        repo_id='boomb0om/dataset-filters',
        filename='watermark_classifier-resnext50_32x4d-input_size320-4epochs_c082_w078.pth',
    )
}


def check_result(ret, testFiles):
    total = len(testFiles)
    correct = 0
    for r, f in zip(ret, testFiles):
        dir_name = os.path.dirname(f)
        if str(r) in dir_name:
            correct += 1
        else:
            pass
            # vprint(f, 'watermark' if r == 1 else 'clear')
    print(correct, total, correct / total)


# https://www.cnblogs.com/wanghui-garcia/p/10679089.html

# https://pytorch.org/tutorials/beginner/finetuning_torchvision_models_tutorial.html
# "https://download.pytorch.org/models/resnext101_32x8d-8ba56ff5.pth" to /home/faith/.cache/torch/hub/checkpoints/resnext101_32x8d-8ba56ff5.pth
def test_and_train():
    device = torch.device('cuda:0')
    # model_name = 'resnext50_32x4d-small'
    model_name = 'resnext101_32x8d-large'

    # bug
    # img = Image.open('/home/faith/aicmder/dl/dataset/1/Picture2.png')
    # img = Image.open('/home/faith/aicmder/dl/Picture5.png')
    # img = Image.open('/home/faith/aicmder/dl/dataset/0/3.jpg')

    # img = Image.open('/home/faith/aicmder/dl/1111.png')
    # predict
    # start = time.time()
    # res = predict_image(img, model, device)
    # end = time.time()
    # print(end - start, 'watermark' if res == 1 else 'clear')

    # testFiles = glob.glob("/home/faith/aicmder/dl/dataset/*/*")
    # testFiles = ["/home/faith/aicmder/dl/dataset/0/IMG_5980.jpg"]

    # predictor = WatermarksPredictor("/home/faith/aicmder/train/best.pt", device, bs=256, verbose=True, name=model_name)
    # ret = predictor.run(testFiles)
    # print(ret)
    # check_result(ret, testFiles)

    print("-----------------------")
    predictor = WatermarksPredictor(None, device, bs=32, verbose=True, name='resnext101_32x8d-large', feature_extract=False)
    # predictor = WatermarksPredictor("/old_faith/home/faith/aicmder/train/best.pt", device, bs=64, verbose=True, name='resnext101_32x8d-large', feature_extract=False)
    criterion = nn.NLLLoss()
    predictor.train("/home/faith/aicmder/dl/dataset", criterion=criterion, epochs=20, opt='Adam', lr=0.125)
    # ret = predictor.run(testFiles)
    # check_result(ret, testFiles)


def test_base64():
    from YoloModule import readb64
    device = torch.device('cuda:0')

    image_file = "/home/faith/aicmder/dl/dataset/0/IMG_5980.jpg"
    image_file = "/home/faith/aicmder/dl/room472.jpg"
    image_file = "/home/faith/aicmder/tests_model/wx1.png"
    image_file = "/home/faith/aicmder/dl/IMG_5980.jpg"
    image_file = "/old_faith/home/faith/aicmder/tests_model/IMG_6006.JPG"
    image_file = "/old_faith/home/faith/aicmder/dl/IMG_6007.JPG"
    # image_file = "/home/faith/aicmder/dl/IMG_6008.JPG"
    # image_file = "/home/faith/aicmder/dl/dataset/1/6.jpg"

    # image_file = "/home/faith/aicmder/dl/dataset/1/Picture5.png"
    # image_file = "/home/faith/aicmder/dl/dataset/1/Picture2.png"
    # image_file = "/old_faith/home/faith/aicmder/dl/dataset/1/Picture3.png"

    # image_file = "/old_faith/home/faith/aicmder/dl/dataset2/1/H2202286_image_20220525140211376.jpg"
    image_file = "/old_faith/home/faith/aicmder/dl/dataset2/1/H2202286_image_20220525140211477.jpg" # not working
    # image_file = "/old_faith/home/faith/aicmder/dl/dataset2/1/H2022653_image_20210615003225318.JPG"
    # image_file = "/old_faith/home/faith/aicmder/dl/dataset2/1/H2022653_image_20210615003225232.JPG"
    with open(image_file,  'rb') as img_f:
        img = img_f.read()
        img_base64 = base64.b64encode(img).decode('utf8')
        img_bgr = readb64(img_base64)
        predictor = WatermarksPredictor(
            "/old_faith/home/faith/aicmder/train/best.pt", device, bs=4, verbose=True, name='resnext101_32x8d-large')
        # predictor = WatermarksPredictor(None, device, bs=32, verbose=True, name='resnext101_32x8d-large')

        # a = glob.glob("/home/faith/wm-nowm/d/*/*")
        # a = glob.glob("/home/faith/aicmder/dl/dataset/*/*")
        print(predictor.run([img_bgr]))

        # ret = predictor.run("/home/faith/wm-nowm/d/")
        # ret = predictor.run("/home/faith/aicmder/dl/dataset2")
        # print(ret)
        # check_result(ret, ret)


if __name__ == "__main__":
    test_and_train()
    # test_base64()
    # print("hello")
    
# OMP_NUM_THREADS=16 python -m torch.distributed.run --nproc_per_node 2 dl/watermark_detection.py    

#### master
# python -m torch.distributed.run --nproc_per_node 2 --nnodes 2 --node_rank 0 --master_addr "192.168.31.130" --master_port 1234 dl/watermark_detection.py

#### slave
# python -m torch.distributed.run --nproc_per_node 1 --nnodes 2 --node_rank 1 --master_addr "192.168.31.130" --master_port 1234 dl/watermark_detection.py 


from .bleu.google_bleu import corpus_bleu
from .rouge.rouge import Rouge
from .meteor.meteor import Meteor
from collections import OrderedDict, Counter


class AverageMeter(object):
    """Computes and stores the average and current value."""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def normalize_text(s):
    """Lower text and remove extra whitespace."""

    def white_space_fix(text):
        return ' '.join(text.split())

    def lower(text):
        return text.lower()

    return white_space_fix(lower(s))


def eval_score(prediction, ground_truth):
    """Compute the geometric mean of precision and recall for answer tokens."""
    precision, recall, f1 = 0, 0, 0
    if len(ground_truth) == 0:
        if len(prediction) == 0:
            precision, recall, f1 = 1, 1, 1
    else:
        prediction_tokens = normalize_text(prediction).split()
        ground_truth_tokens = normalize_text(ground_truth).split()
        common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
        num_same = sum(common.values())
        if num_same != 0:
            precision = 1.0 * num_same / len(prediction_tokens)
            recall = 1.0 * num_same / len(ground_truth_tokens)
            f1 = (2 * precision * recall) / (precision + recall)

    return precision, recall, f1


def compute_eval_score(prediction, ground_truths):
    assert isinstance(prediction, str)
    precision, recall, f1 = 0, 0, 0
    for gt in ground_truths:
        _prec, _rec, _f1 = eval_score(prediction, gt)
        if _f1 > f1:
            precision, recall, f1 = _prec, _rec, _f1
    return precision, recall, f1


def eval_accuracies(hypotheses, references, copy_info=None, sources=None,
                    filename=None, print_copy_info=False, mode='dev'):
    """An unofficial evalutation helper.
     Arguments:
        hypotheses: A mapping from instance id to predicted sequences.
        references: A mapping from instance id to ground truth sequences.
        copy_info: Map of id --> copy information.
        sources: Map of id --> input text sequence.
        filename:
        print_copy_info:
    """
    assert (sorted(references.keys()) == sorted(hypotheses.keys()))

    # Compute BLEU scores
    # bleu_scorer = Bleu(n=4)
    # _, _, bleu = bleu_scorer.compute_score(references, hypotheses, verbose=0)
    # bleu = compute_bleu(references, hypotheses, max_order=4)['bleu']
    # _, bleu, ind_bleu = nltk_corpus_bleu(hypotheses, references)
    _, bleu, ind_bleu = corpus_bleu(hypotheses, references)

    # Compute ROUGE scores
    rouge_calculator = Rouge()
    rouge_l, ind_rouge = rouge_calculator.compute_score(references, hypotheses)

    # Compute METEOR scores
    if mode == 'test':
        try:
            meteor_calculator = Meteor()
            meteor, _ = meteor_calculator.compute_score(references, hypotheses)
        except:
            meteor = 0
    else:
        meteor = 0

    f1 = AverageMeter()
    precision = AverageMeter()
    recall = AverageMeter()

    fw = open(filename, 'w') if filename else None
    for key in references.keys():
        _prec, _rec, _f1 = compute_eval_score(hypotheses[key][0],
                                              references[key])
        precision.update(_prec)
        recall.update(_rec)
        f1.update(_f1)
        if fw:
            if copy_info is not None and print_copy_info:
                prediction = hypotheses[key][0].split()
                pred_i = [word + ' [' + str(copy_info[key][j]) + ']'
                          for j, word in enumerate(prediction)]
                pred_i = [' '.join(pred_i)]
            else:
                pred_i = hypotheses[key]

            logobj = OrderedDict()
            logobj['id'] = key
            if sources is not None:
                logobj['code'] = sources[key]
            logobj['predictions'] = pred_i
            logobj['references'] = references[key][0] if args.print_one_target \
                else references[key]
            logobj['bleu'] = ind_bleu[key]
            logobj['rouge_l'] = ind_rouge[key]
            fw.write(json.dumps(logobj) + '\n')

    if fw:
        fw.close()
    return bleu * 100, rouge_l * 100, meteor * 100, precision.avg * 100, \
        recall.avg * 100, f1.avg * 100


if __name__ == "__main__":
    from aicmder.metrics import eval_accuracies
    sources, hypotheses, references, copy_dict = dict(), dict(), dict(), dict()
    hypotheses[0] = ["hello"]
    references[0] = ["hello"]

    hypotheses[1] = ["world"]
    references[1] = ["wo"]
    print(eval_accuracies(hypotheses, references))



import torch.nn as nn


class FC(nn.Module):
    def __init__(self, inFeat, nbCls=2, dropoutRatio=0.2):
        super(FC, self).__init__()
        # self.dropout = nn.Dropout(dropoutRatio)
        # self.fc1 = nn.Linear(inFeat, nbCls)
        self.sequential = nn.Sequential(nn.Linear(inFeat, 512),
                                        nn.ReLU(),
                                        nn.Dropout(dropoutRatio),
                                        nn.Linear(512, nbCls),
                                        nn.LogSoftmax(dim=1))

    def forward(self, x):
        # x = self.dropout(x)
        # x = self.fc1(x)
        x = self.sequential(x)
        return x


import torch
from torch.utils.data import dataloader
from typing import Any, Callable, Dict
from torch.utils.data import Dataset, Sampler

class DatasetFromSampler(Dataset):
    """Dataset to create indexes from `Sampler`.

    Args:
        sampler: PyTorch sampler
    """

    def __init__(self, sampler: Sampler):
        """Initialisation for DatasetFromSampler."""
        self.sampler = sampler
        self.sampler_list = None

    def __getitem__(self, index: int):
        """Gets element of the dataset.

        Args:
            index: index of the element in the dataset

        Returns:
            Single element by index
        """
        if self.sampler_list is None:
            self.sampler_list = list(self.sampler)
        return self.sampler_list[index]

    def __len__(self) -> int:
        """
        Returns:
            int: length of the dataset
        """
        return len(self.sampler)


class SelfSupervisedDatasetWrapper(Dataset):
    """The Self Supervised Dataset.

    The class implemets contrastive logic (see Figure 2 from `A Simple Framework
    for Contrastive Learning of Visual Representations`_.)

    Args:
        dataset: original dataset for augmentation
        transforms: transforms which will be applied to original batch to get both
            left and right output batch.
        transform_left: transform only for left batch
        transform_right: transform only for right batch
        transform_original: transforms which will be applied to save original in batch
        is_target: the flag for selection does dataset return (sample, target)
            or only sample

    Example:

    .. code-block:: python

        import torchvision
        from torchvision.datasets import CIFAR10

        from catalyst.data.dataset import SelfSupervisedDatasetWrapper

        transforms = torchvision.transforms.Compose(
            [
                torchvision.transforms.RandomResizedCrop(32),
                torchvision.transforms.ToTensor(),
                torchvision.transforms.Normalize(
                    [0.4914, 0.4822, 0.4465], [0.2023, 0.1994, 0.2010]
                ),
                torchvision.transforms.ColorJitter(0.8, 0.8, 0.8, 0.2),
            ]
        )

        cifar_dataset = CIFAR10(root="./data", download=True, transform=None)
        cifar_contrastive = SelfSupervisedDatasetWrapper(
            cifar_dataset,
            transforms=transforms
        )

        for transformed_sample, aug_1, aug_2 in cifar_contrastive:
            pass

    .. _`A Simple Framework for Contrastive Learning of Visual Representations`:
        https://arxiv.org/abs/2002.05709
    """

    def __init__(
        self,
        dataset: Dataset,
        transforms: Callable = None,
        transform_left: Callable = None,
        transform_right: Callable = None,
        transform_original: Callable = None,
        is_target: bool = True,
    ) -> None:
        """
        Args:
            dataset: original dataset for augmentation
            transforms: transforms which will be applied to original batch to get both
            left and right output batch.
            transform_left: transform only for left batch
            transform_right: transform only for right batch
            transform_original: transforms which will be applied
                to save original in batch
            is_target: the flag for selection does dataset return (sample, target)
                or only sample

        Raises:
            ValueError: should be specified transform_left
                and transform_right simultaneously
                or only transforms
        """
        super().__init__()

        if transform_right is not None and transform_left is not None:
            self.transform_right = transform_right
            self.transform_left = transform_left
        elif transforms is not None:
            self.transform_right = transforms
            self.transform_left = transforms
        else:
            raise ValueError(
                "Specify `transform_left` and `transform_right` simultaneously "
                "or only `transforms`."
            )
        self.transform_original = transform_original
        self.dataset = dataset
        self.is_target = is_target

    def __len__(self) -> int:
        """Length"""
        return len(self.dataset)

    def __getitem__(self, idx) -> Dict[str, Any]:
        """Get item method for dataset

        Args:
            idx: index of the object

        Returns:
            Dict with left agumention (aug1), right agumention (aug2) and target
        """
        if self.is_target:
            sample, target = self.dataset[idx]
        else:
            sample = self.dataset[idx]

        transformed_sample = (
            self.transform_original(sample) if self.transform_original else sample
        )
        aug_1 = self.transform_left(sample)
        aug_2 = self.transform_right(sample)

        if self.is_target:
            return transformed_sample, aug_1, aug_2, target
        return transformed_sample, aug_1, aug_2


class InfiniteDataLoader(dataloader.DataLoader):
    """ Dataloader that reuses workers

    Uses same syntax as vanilla DataLoader
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        object.__setattr__(self, 'batch_sampler', _RepeatSampler(self.batch_sampler))
        self.iterator = super().__iter__()

    def __len__(self):
        return len(self.batch_sampler.sampler)

    def __iter__(self):
        for _ in range(len(self)):
            try:
                yield next(self.iterator)
            except Exception as e:
                pass


class _RepeatSampler:
    """ Sampler that repeats forever

    Args:
        sampler (Sampler)
    """

    def __init__(self, sampler):
        self.sampler = sampler

    def __iter__(self):
        while True:
            yield from iter(self.sampler)


class PadCollate:
    """
    a variant of callate_fn that pads according to the longest sequence in
    a batch of sequences
    """

    def __init__(self, dim=0):
        """
        args:
            dim - the dimension to be padded (dimension of time in sequences)
        """
        self.dim = dim

    def pad_tensor(self, vec, pad):
        """
        args:
            vec - tensor to pad
            pad - the size to pad to
            dim - dimension to pad

        return:
            a new tensor padded to 'pad' in dimension 'dim'
        """
        pad_size = list(vec.shape)
        pad_size[self.dim] = pad - vec.size(self.dim)
        return torch.cat([vec, torch.zeros(*pad_size)], dim=self.dim)


    def pad_collate(self, batch):
        raise NotImplementedError('Please implement [pad_collate] method!')

    def __call__(self, batch):
        return self.pad_collate(batch)

__all__ = ["DatasetFromSampler", "SelfSupervisedDatasetWrapper", "InfiniteDataLoader", "PadCollate"]


# !/usr/bin/env python3
"""
==== No Bugs in code, just some Random Unexpected FEATURES ====
┌─────────────────────────────────────────────────────────────┐
│┌───┬───┬───┬───┬───┬───┬───┬───┬───┬───┬───┬───┬───┬───┬───┐│
││Esc│!1 │@2 │#3 │$4 │%5 │^6 │&7 │*8 │(9 │)0 │_- │+= │|\ │`~ ││
│├───┴─┬─┴─┬─┴─┬─┴─┬─┴─┬─┴─┬─┴─┬─┴─┬─┴─┬─┴─┬─┴─┬─┴─┬─┴─┬─┴───┤│
││ Tab │ Q │ W │ E │ R │ T │ Y │ U │ I │ O │ P │{[ │}] │ BS  ││
│├─────┴┬──┴┬──┴┬──┴┬──┴┬──┴┬──┴┬──┴┬──┴┬──┴┬──┴┬──┴┬──┴─────┤│
││ Ctrl │ A │ S │ D │ F │ G │ H │ J │ K │ L │: ;│" '│ Enter  ││
│├──────┴─┬─┴─┬─┴─┬─┴─┬─┴─┬─┴─┬─┴─┬─┴─┬─┴─┬─┴─┬─┴─┬─┴────┬───┤│
││ Shift  │ Z │ X │ C │ V │ B │ N │ M │< ,│> .│? /│Shift │Fn ││
│└─────┬──┴┬──┴──┬┴───┴───┴───┴───┴───┴──┬┴───┴┬──┴┬─────┴───┘│
│      │Fn │ Alt │         Space         │ Alt │Win│   HHKB   │
│      └───┴─────┴───────────────────────┴─────┴───┘          │
└─────────────────────────────────────────────────────────────┘

训练过程中的记录器，类似于SummaryWriter功能，只是SummaryWriter需要依赖于tensorboard和浏览器做可视化，
该工具依赖matplotlib采用静态本地图片存储的形式，便于服务器快速查看训练结果。

Authors: pankeyu
Date: 2021/10/17
"""
import os

import numpy as np
import matplotlib.pyplot as plt


class iSummaryWriter(object):

    def __init__(self, log_path: str, log_name: str, params=[], extention='.png', max_columns=2,
                 log_title=None, figsize=None):
        """
        初始化函数，创建日志类。

        Args:
            log_path (str): 日志存放文件夹
            log_name (str): 日志文件名
            parmas (list): 要记录的参数名字列表，e.g. -> ["loss", "reward", ...]
            extension (str): 图片存储格式
            max_columns (int): 一行中排列几张图，默认为一行2张（2个变量）的图。
        """
        self.log_path = log_path
        if not os.path.exists(log_path):
            os.makedirs(log_path)
        self.log_name = log_name
        self.extention = extention
        self.max_param_index = -1
        self.max_columns_threshold = max_columns
        self.figsize = figsize
        self.params_dict = self.create_params_dict(params)
        self.log_title = log_title
        self.init_plt()
        self.update_ax_list()

    def init_plt(self) -> None:
        plt.style.use('seaborn-darkgrid')

    def create_params_dict(self, params: list) -> dict:
        """
        根据传入需要记录的变量名列表，创建监控变量字典。

        Args:
            params (list): 监控变量名列表

        Returns:
            dict: 监控变量名字典 -> {
                'loss': {'values': [0.44, 0.32, ...], 'epochs': [10, 20, ...], 'index': 0},
                'reward': {'values': [10.2, 13.2, ...], 'epochs': [10, 20, ...], 'index': 1},
                ...
            }
        """
        params_dict = {}
        for i, param in enumerate(params):
            params_dict[param] = {'values': [], 'epochs': [], 'index': i}
            self.max_param_index = i
        return params_dict

    def update_ax_list(self) -> None:
        """
        根据当前的监控变量字典，为每一个变量分配一个图区。
        """
        # * 重新计算每一个变量对应的图幅索引
        params_num = self.max_param_index + 1
        if params_num <= 0:
            return

        self.max_columns = params_num if params_num < self.max_columns_threshold else self.max_columns_threshold
        max_rows = (params_num - 1) // self.max_columns + 1   # * 所有变量最多几行
        figsize = self.figsize if self.figsize else (self.max_columns * 6,max_rows * 3)    # 根据图个数计算整个图的figsize
        self.fig, self.axes = plt.subplots(max_rows, self.max_columns, figsize=figsize)

        # * 如果只有一行但又不止一个图，需要手动reshape成(1, n)的形式
        if params_num > 1 and len(self.axes.shape) == 1:
            self.axes = np.expand_dims(self.axes, axis=0)

        # * 重新设置log标题
        log_title = self.log_title if self.log_title else '[Training Log] {}'.format(
            self.log_name)
        self.fig.suptitle(log_title, fontsize=15)

    def add_scalar(self, param: str, value: float, epoch: int) -> None:
        """
        添加一条新的变量值记录。

        Args:
            param (str): 变量名，e.g. -> 'loss'
            value (float): 此时的值。
            epoch (int): 此时的epoch数。
        """
        # * 如果该参数是第一次加入，则将该参数加入到监控变量字典中
        if param not in self.params_dict:
            self.max_param_index += 1
            self.params_dict[param] = {'values': [],
                                       'epochs': [], 'index': self.max_param_index}
            self.update_ax_list()

        self.params_dict[param]['values'].append(value)
        self.params_dict[param]['epochs'].append(epoch)

    def record(self, dpi=200) -> None:
        """
        调用该接口，对该类中目前所有监控的变量状态进行一次记录，将结果保存到本地文件中。
        """
        for param, param_elements in self.params_dict.items():
            param_index = param_elements["index"]
            param_row, param_column = param_index // self.max_columns, param_index % self.max_columns
            ax = self.axes[param_row, param_column] if self.max_param_index > 0 else self.axes
            # ax.set_title(param)
            ax.set_xlabel('Epoch')
            ax.set_ylabel(param)
            ax.plot(self.params_dict[param]['epochs'],
                    self.params_dict[param]['values'],
                    color='darkorange')

        plt.savefig(os.path.join(self.log_path,
                    self.log_name + self.extention), dpi=dpi)


if __name__ == '__main__':
    import random
    import time

    n_epochs = 10
    log_path, log_name = './', 'test'
    writer = iSummaryWriter(log_path=log_path, log_name=log_name)
    for i in range(n_epochs):
        loss, reward = 100 - random.random() * i, random.random() * i
        writer.add_scalar('loss', loss, i)
        writer.add_scalar('reward', reward, i)
        writer.add_scalar('random', reward, i)
        writer.record()
        print("Log has been saved at: {}".format(
            os.path.join(log_path, log_name)))
        time.sleep(3)


import nltk
from nltk.translate.bleu_score import SmoothingFunction


def nltk_sentence_bleu(hypothesis, reference, order=4):
    cc = SmoothingFunction()
    return nltk.translate.bleu([reference], hypothesis, smoothing_function=cc.method4)


def nltk_corpus_bleu(hypotheses, references, order=4):
    refs = []
    hyps = []
    count = 0
    total_score = 0.0

    cc = SmoothingFunction()

    assert (sorted(hypotheses.keys()) == sorted(references.keys()))
    Ids = list(hypotheses.keys())
    ind_score = dict()

    for id in Ids:
        hyp = hypotheses[id][0].split()
        ref = [r.split() for r in references[id]]
        hyps.append(hyp)
        refs.append(ref)

        score = nltk.translate.bleu(ref, hyp, smoothing_function=cc.method4)
        total_score += score
        count += 1
        ind_score[id] = score

    avg_score = total_score / count
    corpus_bleu = nltk.translate.bleu_score.corpus_bleu(refs, hyps)
    return corpus_bleu, avg_score, ind_score


if __name__ == "__main__":
    sources, hypotheses, references, copy_dict = dict(), dict(), dict(), dict()
    hypotheses[0] = ["hello"]
    references[0] = ["hello"]

    hypotheses[1] = ["world"]
    references[1] = ["ow"]
    print(nltk_corpus_bleu(hypotheses, references))


import logging
import os
import torch.nn as nn
import torch

VERBOSE = str(os.getenv('TRAIN_VERBOSE', True)
              ).lower() == 'true'  # global verbose mode


def set_logging(name=None, verbose=VERBOSE):
    # Sets level and returns logger
    # if is_kaggle():
    #     for h in logging.root.handlers:
    #         logging.root.removeHandler(h)  # remove all handlers associated with the root logger object
    rank = int(os.getenv('RANK', -1))  # rank in world for Multi-GPU trainings
    level = logging.INFO if verbose and rank in {-1, 0} else logging.ERROR
    log = logging.getLogger(name)
    log.setLevel(level)
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(message)s"))
    handler.setLevel(level)
    log.addHandler(handler)


def colorstr(*input):
    # Colors a string https://en.wikipedia.org/wiki/ANSI_escape_code, i.e.  colorstr('blue', 'hello world')
    # color arguments, string
    *args, string = input if len(input) > 1 else ('blue', 'bold', input[0])
    colors = {
        'black': '\033[30m',  # basic colors
        'red': '\033[31m',
        'green': '\033[32m',
        'yellow': '\033[33m',
        'blue': '\033[34m',
        'magenta': '\033[35m',
        'cyan': '\033[36m',
        'white': '\033[37m',
        'bright_black': '\033[90m',  # bright colors
        'bright_red': '\033[91m',
        'bright_green': '\033[92m',
        'bright_yellow': '\033[93m',
        'bright_blue': '\033[94m',
        'bright_magenta': '\033[95m',
        'bright_cyan': '\033[96m',
        'bright_white': '\033[97m',
        'end': '\033[0m',  # misc
        'bold': '\033[1m',
        'underline': '\033[4m'}
    return ''.join(colors[x] for x in args) + f'{string}' + colors['end']


set_logging()  # run before defining LOGGER
# define globally (used in train.py, val.py, detect.py, etc.)
LOGGER = logging.getLogger("train")


def smart_optimizer(model, name='Adam', lr=0.001, momentum=0.9, weight_decay=1e-5):
    # YOLOv5 3-param group optimizer: 0) weights with decay, 1) weights no decay, 2) biases no decay
    g = [], [], []  # optimizer parameter groups
    # normalization layers, i.e. BatchNorm2d()
    bn = tuple(v for k, v in nn.__dict__.items() if 'Norm' in k)
    for v in model.modules():
        if hasattr(v, 'bias') and isinstance(v.bias, nn.Parameter):  # bias (no decay)
            g[2].append(v.bias)
        if isinstance(v, bn):  # weight (no decay)
            g[1].append(v.weight)
        elif hasattr(v, 'weight') and isinstance(v.weight, nn.Parameter):  # weight (with decay)
            g[0].append(v.weight)

    if name == 'Adam':
        optimizer = torch.optim.Adam(g[2], lr=lr, betas=(
            momentum, 0.999))  # adjust beta1 to momentum
    elif name == 'AdamW':
        optimizer = torch.optim.AdamW(
            g[2], lr=lr, betas=(momentum, 0.999), weight_decay=0.0)
    elif name == 'RMSProp':
        optimizer = torch.optim.RMSprop(g[2], lr=lr, momentum=momentum)
    elif name == 'SGD':
        optimizer = torch.optim.SGD(
            g[2], lr=lr, momentum=momentum, nesterov=True)
    else:
        raise NotImplementedError(f'Optimizer {name} not implemented.')

    # add g0 with weight_decay
    optimizer.add_param_group({'params': g[0], 'weight_decay': weight_decay})
    # add g1 (BatchNorm2d weights)
    optimizer.add_param_group({'params': g[1], 'weight_decay': 0.0})
    LOGGER.info(f"{colorstr('optimizer:')} {type(optimizer).__name__} with parameter groups "
                f"{len(g[1])} weight (no decay), {len(g[0])} weight, {len(g[2])} bias")
    return optimizer



from aicmder.torch import HyperParameters, DataModule, select_device
from aicmder.torch.dist_model import *
from torch.optim import lr_scheduler
from pathlib import Path
import torch.nn as nn
import torch
from copy import deepcopy
import pandas as pd
import matplotlib.pyplot as plt
from aicmder.utils.general import multiple_threaded
from matplotlib.ticker import MaxNLocator
try:

    import seaborn as sns
    sns.set_context("paper")

    bg_color = "#f1f1f2"
    grid_color = "#bcbabe"
    text_color = "#338494"
    tick_color = "#1995ad"

    sns.set(rc={
        #  'axes.axisbelow': False,
        'axes.edgecolor': 'lightgrey',
        'axes.facecolor': 'None',  # remove bg
        #  'axes.grid': False,
        'axes.labelcolor': 'dimgrey',
        #  'axes.spines.right': False,
        #  'axes.spines.top': False,
        'figure.facecolor': 'white',
        'lines.solid_capstyle': 'round',
        'patch.edgecolor': 'w',
        'patch.force_edgecolor': True,
        'text.color': 'dimgrey',
        #  'xtick.bottom': False,
        'xtick.color': 'dimgrey',
        'xtick.direction': 'out',
        #  'xtick.top': False,
        'ytick.color': 'dimgrey',
        'ytick.direction': 'out',
        #  'ytick.left': False,
        #  'ytick.right': False


        # add background
        'axes.facecolor': bg_color,
        'figure.facecolor': bg_color,
        "text.color": text_color,
        "xtick.color": tick_color,
        "ytick.color": tick_color,
        "axes.edgecolor": grid_color,

        "axes.labelcolor": text_color,
        "grid.color": grid_color,
    })
except:
    pass
from matplotlib import rcParams
rcParams['figure.figsize'] = 13, 8.27

CB91_Blue = '#2CBDFE'
CB91_Green = '#47DBCD'
CB91_Pink = '#F3A0F2'
CB91_Purple = '#9D2EC5'
CB91_Violet = '#661D98'
CB91_Amber = '#F5B14C'
color_list = [CB91_Blue, CB91_Pink, CB91_Green, CB91_Amber,
              CB91_Purple, CB91_Violet]


def write_frame(*p):
    d = p[0]
    temp_df = pd.read_csv(d['filename'])

    col_len = temp_df.columns.size
    columns = temp_df.columns
    # exclude epoch
    sub_size = math.ceil(math.sqrt(col_len - 1))
    
    df_idx = p[1]
    fig, axs = plt.subplots(nrows=sub_size, ncols=sub_size)
    idx, need_stop = 1, False
    for i in range(sub_size):
        for j in range(sub_size):
            dfm = temp_df[:df_idx].iloc[:, [0, idx]]
            metric = columns[idx]
            dfm.columns = ["epoch", metric]
            ax = axs[i][j]
            ax.title.set_text(metric)
            color_idx = idx % len(color_list)

            ax.xaxis.set_major_locator(MaxNLocator(integer=True))
            plt.tight_layout()
            sns.lineplot(x="epoch", y=metric, data=dfm, ax=ax, color=color_list[color_idx])

            idx += 1
            if idx >= col_len:
                need_stop = True
                break
        if need_stop:
            break
    plt.tight_layout()
    df_filename = f"/tmp/{df_idx}.png"
    plt.savefig(df_filename)
    # files.append(df_filename)
    return df_filename


_STA_CSV = 'statis.csv'

_STA_VALIDATING = 0
_STA_TRAINING = 1


class Trainer(HyperParameters):

    def __init__(self, model, device='', max_epochs=100, batch_size=32, gradient_clip_val=10.0, verbose=True,
                 evaluate_func=evaluate_func, patience=5, optim=None, save_dir="./", save_state_dict_only=True,
                 lrf=0.01, use_lr_scheduler=True, save_statistic=False,
                 criterion=nn.CrossEntropyLoss()):
        self.save_hyperparameters()
        self.trainable = Trainable(batch_size, gradient_clip_val=self.gradient_clip_val)
        self.trainable.on_pretrain_routine_start(self.model, verbose=verbose)
        self.device = select_device(device=device, batch_size=batch_size)
        if hasattr(model, "_get_name"):
            self.model_name = model._get_name()
        else:
            self.model_name = ''

        self.best_pt_name = self.model_name + 'best.pt'
        self.last_pt_name = self.model_name + 'last.pt' 

        self.model, self.device, self.amp, self.scaler, self.ema = self.trainable.on_train_start(
            model=self.model, device=self.device, evaluate_func=evaluate_func, patience=patience)
        save_dir = Path(save_dir)
        print("--------------------------------------------------------------------------------------")
        print("warning:", "AMP [{}]".format(self.amp))
        print("--------------------------------------------------------------------------------------")
        self.weight_dir = save_dir / 'train'  # weights dir
        self.weight_dir.mkdir(parents=True, exist_ok=True)  # make dir
        self.last, self.best = self.weight_dir / self.last_pt_name, self.weight_dir / self.best_pt_name
        self.val_save = False
        if self.optim is not None:
            if use_lr_scheduler:
                def lr_lambda(x): return (1 - x / self.max_epochs) * (1.0 - lrf) + lrf  # linear
                self.scheduler = lr_scheduler.LambdaLR(self.optim, lr_lambda=lr_lambda)
            else:
                self.scheduler = self.optim

    def to_csv(self):
        if self.save_statistic:
            self.df.to_csv(self.weight_dir / self.statis_csv, index=False)

    def save_statistic_epoch(self):
        if self.save_statistic:
            self.setValueByRowIdx(self.epoch, self.df_epoch, self.epoch)
            self.setValueByRowIdx(self.epoch, self.df_acc, self.epoch_acc)
            self.setValueByRowIdx(self.epoch, self.df_loss, self.epoch_loss)

            if hasattr(self, "need_val") and self.need_val:
                self.setValueByRowIdx(self.epoch, self.df_val_acc, self.val_acc)
                self.setValueByRowIdx(self.epoch, self.df_val_loss, self.val_loss)

    def setValueByRowIdx(self, row_idx, key, val):
        if self.save_statistic:
            if isinstance(val, torch.Tensor):
                val = '%.3g' % val.item()
            self.df.at[row_idx, key] = val

    def create_dataframe(self):
        self.df_epoch = 'epoch'
        self.df_acc = 'acc'
        self.df_loss = 'loss'

        self.df_val_acc = 'val_acc'
        self.df_val_loss = 'val_loss'

        self.df[self.df_epoch] = ''
        self.df[self.df_acc] = ''
        self.df[self.df_loss] = ''

        self.df[self.df_val_acc] = ''
        self.df[self.df_val_loss] = ''

    def configure_statistic(self):
        if self.save_statistic:
            import pandas as pd
            self.df = pd.DataFrame()
            self.create_dataframe()

    def configure_criterion(self):
        pass

    def configure_optimizers(self, opt='SGD',
                             #  opt='Adam',
                             lr=0.01,
                             lrf=0.01,
                             momentum=0.937,
                             weight_decay=0.0005):
        self.optim = smart_optimizer(self.model, opt, lr, momentum, weight_decay)
        if self.use_lr_scheduler:
            def lr_lambda(x): return (1 - x / self.max_epochs) * (1.0 - lrf) + lrf  # linear
            self.scheduler = lr_scheduler.LambdaLR(self.optim, lr_lambda=lr_lambda)
        else:
            self.scheduler = self.optim

    def prepare_data(self, data):
        assert isinstance(data, DataModule)
        self.train_dataloader = data.train_dataloader()
        self.val_dataloader = data.val_dataloader()
        self.num_train_batches = len(self.train_dataloader)
        self.num_val_batches = (len(self.val_dataloader)
                                if self.val_dataloader is not None else 0)

    def clip_gradients(self, grad_clip_val, model):
        """Defined in :numref:`sec_rnn-scratch`"""
        params = [p for p in model.parameters() if p.requires_grad]
        norm = torch.sqrt(sum(torch.sum((p.grad ** 2)) for p in params))
        if norm > grad_clip_val:
            for param in params:
                param.grad[:] *= grad_clip_val / norm

    def get_current_metrics(self):
        if not hasattr(self, 'best_train_acc'):
            self.best_train_acc = self.epoch_acc

        if not hasattr(self, 'best_train_loss'):
            self.best_train_loss = self.epoch_loss

        if self.epoch_acc > self.best_train_acc or (
                self.epoch_acc == self.best_train_acc and self.best_train_loss is not None and self.epoch_loss < self.best_train_loss):
            self.best_train_acc = self.epoch_acc
            self.best_train_loss = self.epoch_loss
            self.trainable.on_model_save(self.model, self.epoch, self.optim, (self.best_train_acc,
                                        self.best_train_loss), self.best, save_state_dict_only=True)

        return self.best_train_acc, self.best_train_loss
        # raise NotImplementedError("please rewrite this method!")
    #    return (best_train_acc, best_train_loss)

    def run_model(self, batch, is_train=True):
        raise NotImplementedError("please rewrite [run_model] method!")

    def eval_model(self, batch):
        raise NotImplementedError("please rewrite [eval_model] method!")

    def to_device(self, batch):
        if isinstance(batch, torch.Tensor):
            batch = batch.to(self.device)
        if isinstance(batch, list) or isinstance(batch, tuple):
            device_batch = []
            for b in batch:
                if isinstance(b, torch.Tensor):
                    b = b.to(self.device)
                    device_batch.append(b)
            batch = device_batch
        return batch

    def print_val_metrics(self):
        pass

    def is_training(self):
        return self.status == _STA_TRAINING

    def eval(self, data):
        self.prepare_data(data)
        self.model.eval()
        self.epoch = 0
        self.max_epochs = 0
        self.status = _STA_VALIDATING
        if RANK in {-1, 0}:
            self.val_count_in_epoch = len(self.val_dataloader.dataset)
            self.val_corrects_in_epoch = 0

            val_pbar = enumerate(self.val_dataloader)
            nb = len(self.val_dataloader)  # number of batches
            val_pbar = tqdm(val_pbar, total=nb, bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}')
            for idx, batch in val_pbar:
                with torch.no_grad():
                    batch = self.to_device(batch)
                    self.eval_model(batch)

                    self.val_acc = self.val_corrects_in_epoch / self.val_count_in_epoch if self.val_count_in_epoch != 0 else 0
                    try:
                        mem = f'{torch.cuda.memory_reserved() / 1E9 if torch.cuda.is_available() else 0:.3g}G'  # (GB)
                        val_desc = (
                            '%15s' * 1 + '%15s' * 3) % (f'Mem:{mem}',
                                                        "Correct: {}".format(self.val_corrects_in_epoch),
                                                        "Cnt: {}".format(self.val_count_in_epoch),
                                                        "Acc: {:2f}".format(self.val_acc))
                    except Exception as e:
                        val_desc = ''
                    val_desc = 'Validate:' + val_desc
                    val_pbar.set_description(val_desc)

            # for
            self.print_val_metrics()


    def train(self, data, need_val=True, filename=_STA_CSV, save_png_epoch_interval=0):
        if self.optim is None:
            self.configure_optimizers()

        if data is not None and type(data).__name__ != '':
            self.dataset_name = type(data).__name__


        self.status = _STA_TRAINING
        self.need_val = need_val

        if filename == _STA_CSV: 
            pass
        else:
            filename, _ = os.path.splitext(filename)

        self.prefix = filename
        self.statis_csv = f"{filename}{self.dataset_name}_{self.model_name}.csv"
        self.statis_fig = f"{filename}{self.dataset_name}_{self.model_name}.png"
        self.statis_gif = f"{filename}{self.dataset_name}_{self.model_name}.gif"

        if self.statis_csv != _STA_CSV:
            self.last, self.best = self.weight_dir / (filename + f"{self.dataset_name}_" + self.last_pt_name), self.weight_dir / (filename +  f"{self.dataset_name}_" + self.best_pt_name)

        self.configure_statistic()
        self.configure_criterion()
        self.prepare_data(data)

        # self.optim = model.configure_optimizers()
        self.epoch = 0
        self.train_batch_idx = 0
        self.val_batch_idx = 0
        for self.epoch in range(self.max_epochs):

            if RANK in {-1, 0}:
                self.train_loss_in_epoch = 0.0
                self.train_corrects_in_epoch = 0
                self.train_count_in_epoch = len(self.train_dataloader.dataset)

            self.train_epoch()
            # epoch end

            if RANK in {-1, 0}:
                self.epoch_loss = self.train_loss_in_epoch / self.train_count_in_epoch
                self.epoch_acc = self.train_corrects_in_epoch.double() / self.train_count_in_epoch

                print('training {} Loss: {:.4f} Acc: {:.4f} Count: {} Train_size: {}'.format(
                    self.epoch, self.epoch_loss, self.epoch_acc, self.train_count_in_epoch, self.train_size))

                self.save_statistic_epoch()

            stop = self.trainable.on_train_epoch_end(self.epoch, self.get_current_metrics(), scheduler=self.scheduler)
            if stop:
                break
            if self.save_statistic and save_png_epoch_interval > 0:
                if self.epoch % save_png_epoch_interval == 0:
                    self.save_fig()

        self.trainable.on_model_save(self.model, self.epoch, self.optim, self.get_current_metrics(),
                                     self.last, save_state_dict_only=self.save_state_dict_only)
        self.trainable.on_train_end()
        self.to_csv()

    def train_epoch(self):
        """Defined in :numref:`sec_linear_scratch`"""
        # self.model.train()
        pbar, self.train_size = self.trainable.on_train_epoch_start(self.model, self.epoch, self.train_dataloader)
        for batch_idx, batch in pbar:
            ni = batch_idx + self.epoch * self.train_size

            batch = self.to_device(batch)
            # loss = self.model.training_step(self.prepare_batch(batch))
            with torch.cuda.amp.autocast(self.amp):
                loss, desc = self.run_model(batch)

            try:
                mem = f'{torch.cuda.memory_reserved() / 1E9 if torch.cuda.is_available() else 0:.3g}G'  # (GB)
                desc = desc + (
                    '%15s' * 2 + '%15s' * 4) % (
                    f'{self.epoch}/{self.max_epochs - 1}', f'Mem:{mem}', "Loss: {:.2f}".format(self.train_loss_in_epoch),
                    "Correct: {}".format(self.train_corrects_in_epoch),
                    "Cnt: {}".format(self.train_count_in_epoch),
                    "Acc: {:2f}".format(
                        self.train_corrects_in_epoch / self.train_count_in_epoch if self.train_count_in_epoch != 0 else 0))
            except Exception as e:
                desc = ''

            if RANK != -1:
                loss *= WORLD_SIZE  # gradient averaged between devices in DDP mode
            self.trainable.on_train_batch_end(self.model, self.scaler, self.optim, loss, desc=desc, ni=ni)

            # self.optim.zero_grad()
            # with torch.no_grad():
            #     loss.backward()
            #     if self.gradient_clip_val > 0:  # To be discussed later
            #         self.clip_gradients(self.gradient_clip_val, self.model)
            #     self.optim.step()

            self.train_batch_idx += 1
        # eval
        if self.need_val:
            if self.val_dataloader is None:
                return
            self.model.eval()
            if RANK in {-1, 0}:
                self.val_count_in_epoch = len(self.val_dataloader.dataset)
                self.val_loss_in_epoch = 0.0
                self.val_corrects_in_epoch = 0

                val_pbar = enumerate(self.val_dataloader)
                nb = len(self.val_dataloader)  # number of batches
                val_pbar = tqdm(val_pbar, total=nb, bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}')
                for idx, batch in val_pbar:
                    with torch.no_grad():
                        # self.model.validation_step(self.prepare_batch(batch))
                        batch = self.to_device(batch)
                        self.run_model(batch, is_train=False)
                        if self.val_save:
                            self.trainable.on_model_save(
                                self.model, self.epoch, self.optim, self.get_current_metrics(),
                                self.best, save_state_dict_only=self.save_state_dict_only)

                        self.val_acc = self.val_corrects_in_epoch / self.val_count_in_epoch if self.val_count_in_epoch != 0 else 0
                        try:
                            mem = f'{torch.cuda.memory_reserved() / 1E9 if torch.cuda.is_available() else 0:.3g}G'  # (GB)
                            val_desc = (
                                '%15s' * 1 + '%15s' * 4) % (f'Mem:{mem}',
                                                            "Loss: {:.2f}".format(self.val_loss_in_epoch),
                                                            "Correct: {}".format(self.val_corrects_in_epoch),
                                                            "Cnt: {}".format(self.val_count_in_epoch),
                                                            "Acc: {:2f}".format(self.val_acc))
                        except Exception as e:
                            val_desc = ''
                        val_desc = 'Validate:' + val_desc
                        val_pbar.set_description(val_desc)
                    self.val_batch_idx += 1
                
                # for
                self.print_val_metrics()

                self.val_loss = self.val_loss_in_epoch

    def save_fig(self, debug=False, csv_file=_STA_CSV):
        if debug:
            self.save_statistic = True
            self.df = pd.read_csv(self.weight_dir / csv_file)

        if self.save_statistic:

            import importlib
            _loader = importlib.find_loader('seaborn')
            found = _loader is not None
            if found:
                self.to_csv()

                temp_df = pd.read_csv(self.weight_dir / self.statis_csv)

                col_len = temp_df.columns.size
                columns = temp_df.columns
                # exclude epoch
                sub_size = math.ceil(math.sqrt(col_len - 1))

                fig, axs = plt.subplots(nrows=sub_size, ncols=sub_size)
                idx, need_stop = 1, False
                for i in range(sub_size):
                    for j in range(sub_size):
                        # dfm = deepcopy(self.df.iloc[:, [0, idx]])
                        dfm = temp_df.iloc[:, [0, idx]]
                        metric = columns[idx]
                        dfm.columns = ["epoch", metric]
                        ax = axs[i][j]
                        ax.title.set_text(metric)
                        color_idx = idx % len(color_list)
                        # axs[i][j].plot(dfm["epoch"], dfm[metric], color=color_list[color_idx])
                        # sns.pointplot(x="epoch", y=metric, data=dfm, ax=ax, color=color_list[color_idx])

                        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
                        plt.tight_layout()
                        sns.lineplot(x="epoch", y=metric, data=dfm, ax=ax, color=color_list[color_idx])

                        idx += 1
                        if idx >= col_len:
                            need_stop = True
                            break
                    if need_stop:
                        break
                plt.tight_layout()
                plt.savefig(self.weight_dir / self.statis_fig)
            else:
                print("[Warning]: seaborn not found! Can not save plot!")

    def save_gif(self, debug=False, csv_file=_STA_CSV, last_frame_repeat=40):
        if debug:
            self.save_statistic = True
            self.df = pd.read_csv(self.weight_dir / csv_file)

        self.to_csv()
        temp_df = pd.read_csv(self.statis_csv)

        col_len = temp_df.columns.size
        columns = temp_df.columns
        # exclude epoch
        sub_size = math.ceil(math.sqrt(col_len - 1))

        # files = []
        length = len(temp_df)

        param = {'filename': self.statis_csv}
        files = multiple_threaded(write_frame, param, range(1, length + 1))
        # from tqdm import tqdm
        # for df_idx in tqdm(range(1, length + 1), total=length):

        #     fig, axs = plt.subplots(nrows=sub_size, ncols=sub_size)
        #     idx = 1
        #     for i in range(sub_size):
        #         for j in range(sub_size):
        #             dfm = temp_df[:df_idx].iloc[:, [0, idx]]
        #             metric = columns[idx]
        #             dfm.columns = ["epoch", metric]
        #             ax = axs[i][j]
        #             ax.title.set_text(metric)
        #             color_idx = idx % len(color_list)
        #             plt.tight_layout()
        #             sns.lineplot(x="epoch", y=metric, data=dfm, ax=ax, color=color_list[color_idx])

        #             idx += 1
        #             if idx > col_len:
        #                 break
        #     plt.tight_layout()
        #     df_filename = f"/tmp/{df_idx}.png"
        #     plt.savefig(df_filename)
        #     files.append(df_filename)

        import imageio

        for _ in range(last_frame_repeat):
            files.append(files[-1])
        with imageio.get_writer(self.weight_dir / self.statis_gif, mode='I') as writer:
            for f in files:
                image = imageio.imread(f)
                writer.append_data(image)
